import os
import shutil
from qbitcoin.core import config
from qbitcoin.core.misc import db, logger
from qbitcoin.core.State import State
from qbitcoin.core.Block import Block


# TODO: State Migration integration pending
class StateMigration:
    def __init__(self):
        pass

    def state_migration_step_1(self, state: State) -> bool:
        """
        Migration Step from State Version 0 to 1
        :return:
        """
        if state.is_older_state_version():
            db_dir_v1 = os.path.join(config.user.data_dir, config.dev.db_name + '2')
            self._tmp_state = State(state._db)  # DB Pointing to Older State
            state._db = db.DB(db_dir_v1)  # DB Pointing to Newer State
            return True
        return False

    def height_from_state_version_0(self) -> int:
        return self._tmp_state.get_mainchain_height()

    def block_from_state_version_0(self, block_number):
        return Block.get_block_by_number(self._tmp_state, block_number)

    def state_migration_step_2(self, state: State):
        """
        Migration Step from State Version 0 to 1
        :return:
        """
        del self._tmp_state
        self._tmp_state = None
        del state._db

        tmp_db_dir = os.path.join(config.user.data_dir, config.dev.db_name + "3")
        db_dir = os.path.join(config.user.data_dir, config.dev.db_name)
        shutil.move(db_dir,
                    tmp_db_dir)
        tmp_db_dir = os.path.join(config.user.data_dir, config.dev.db_name + "2")
        shutil.move(tmp_db_dir,
                    db_dir)
        state._db = db.DB()
        logger.warning("State Migration Finished")



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
import sys
import threading
from collections import OrderedDict
from typing import Optional, Tuple
from math import ceil

import functools
from pyqrllib.pyqrllib import bin2hstr
from pyqryptonight.pyqryptonight import StringToUInt256, UInt256ToString

from qbitcoin.core import config, BlockHeader
from qbitcoin.core.config import DevConfig
from qbitcoin.core.formulas import block_reward
from qbitcoin.core.StateContainer import StateContainer
from qbitcoin.core.StateMigration import StateMigration
from qbitcoin.core.AddressState import AddressState
from qbitcoin.core.VoteStats import VoteStats
from qbitcoin.core.LastTransactions import LastTransactions
from qbitcoin.core.TransactionMetadata import TransactionMetadata
from qbitcoin.core.OptimizedAddressState import OptimizedAddressState
from qbitcoin.core.PaginatedData import PaginatedData
from qbitcoin.core.PaginatedBitfield import PaginatedBitfield
from qbitcoin.core.Indexer import Indexer
from qbitcoin.core.Block import Block
from qbitcoin.core.BlockMetadata import BlockMetadata
from qbitcoin.core.DifficultyTracker import DifficultyTracker
from qbitcoin.core.GenesisBlock import GenesisBlock
from qbitcoin.core.MultiSigAddressState import MultiSigAddressState
from qbitcoin.core.PoWValidator import PoWValidator
from qbitcoin.core.txs.Transaction import Transaction
from qbitcoin.core.txs.CoinBase import CoinBase
from qbitcoin.core.txs.LatticeTransaction import LatticeTransaction
from qbitcoin.core.txs.SlaveTransaction import SlaveTransaction
from qbitcoin.core.txs.TransferTokenTransaction import TransferTokenTransaction
from qbitcoin.core.txs.multisig.MultiSigVote import MultiSigVote
from qbitcoin.core.TransactionPool import TransactionPool
from qbitcoin.core.misc import logger
from qbitcoin.crypto.Qryptonight import Qryptonight
from qbitcoin.generated import qbit_pb2, qbitstateinfo_pb2


class ChainManager:
    def __init__(self, state):
        self._state = state
        self.tx_pool = TransactionPool(None)
        self._last_block = Block.deserialize(GenesisBlock().serialize())
        self.current_difficulty = StringToUInt256(str(config.user.genesis_difficulty))

        self.trigger_miner = False
        self.lock = threading.RLock()

    @property
    def re_org_limit(self):
        with self.lock:
            return self._state.get_re_org_limit()

    @property
    def height(self):
        with self.lock:
            if not self._last_block:
                return -1
            return self._last_block.block_number

    @property
    def last_block(self) -> Block:
        with self.lock:
            return self._last_block

    @property
    def total_coin_supply(self):
        with self.lock:
            return self._state.total_coin_supply

    def get_cumulative_difficulty(self):
        with self.lock:
            last_block_metadata = BlockMetadata.get_block_metadata(self._state,
                                                                   self._last_block.headerhash)
            return last_block_metadata.cumulative_difficulty

    def get_block_by_number(self, block_number) -> Optional[Block]:
        with self.lock:
            return Block.get_block_by_number(self._state, block_number)

    def get_block_header_hash_by_number(self, block_number) -> Optional[bytes]:
        with self.lock:
            return Block.get_block_header_hash_by_number(self._state,
                                                         block_number)

    def get_block(self, header_hash: bytes) -> Optional[Block]:
        with self.lock:
            return Block.get_block(self._state, header_hash)

    def get_address_balance(self, address: bytes) -> int:
        with self.lock:
            return self.get_optimized_address_state(address).balance

    def get_address_is_used(self, address: bytes) -> bool:
        with self.lock:
            return self._state.get_address_is_used(address)

    def get_address_state(self,
                          address: bytes,
                          exclude_ots_bitfield: bool = False,
                          exclude_transaction_hashes: bool = False) -> AddressState:
        """
        Transform Optimized Address State into Older Address State format
        This should only be used by API.
        Note: OTS tracking removed for Falcon signature support.
        """
        optimized_address_state = self.get_optimized_address_state(address)
        ots_bitfield = []  # Empty bitfield for Falcon signatures
        transaction_hashes = list()
        tokens = OrderedDict()
        slave_pks_access_type = OrderedDict()

        # OTS tracking not needed for Falcon signatures - skip bitfield processing

        if not exclude_transaction_hashes:
            max_transaction_hash_page = ceil(optimized_address_state.transaction_hash_count() / config.dev.data_per_page)

            for page in range(0, max_transaction_hash_page + 1):
                page_data = self.get_transaction_hashes(address, page * config.dev.data_per_page)
                transaction_hashes.extend(page_data)

            max_token_page = ceil(optimized_address_state.tokens_count() / config.dev.data_per_page)

            for page in range(0, max_token_page + 1):
                page_data = self.get_token_transaction_hashes(address, page * config.dev.data_per_page)
                for token_txn_hash in page_data:
                    token_balance = self.get_token(address, token_txn_hash)
                    # token_balance None is only possible when the token transaction
                    # is done by a QRL address as an owner, which has not been
                    # assigned any token balance.
                    if token_balance is None:
                        continue
                    tokens[token_txn_hash] = token_balance.balance

            max_slave_page = ceil(optimized_address_state.slaves_count() / config.dev.data_per_page)

            for page in range(0, max_slave_page + 1):
                page_data = self.get_slave_transaction_hashes(address, page * config.dev.data_per_page)
                for slave_txn_hash in page_data:
                    tx, _ = self.get_tx_metadata(slave_txn_hash)
                    for slave_pk in tx.slave_pks:
                        slave_meta_data = self.get_slave_pk_access_type(address, slave_pk)
                        slave_pks_access_type[str(slave_pk)] = slave_meta_data.access_type

        addr_state = AddressState.create(address=optimized_address_state.address,
                                         nonce=optimized_address_state.nonce,
                                         balance=optimized_address_state.balance,
                                         falcon_pk_list=[],  # Empty list since using Falcon signatures
                                         tokens=tokens,
                                         slave_pks_access_type=slave_pks_access_type)
        addr_state.transaction_hashes.extend(transaction_hashes)

        return addr_state

    def get_optimized_address_state(self, address: bytes) -> OptimizedAddressState:
        with self.lock:
            return OptimizedAddressState.get_optimized_address_state(self._state, address)

    def get_multi_sig_address_state(self, address: bytes) -> MultiSigAddressState:
        with self.lock:
            return MultiSigAddressState.get_multi_sig_address_state_by_address(self._state._db, address)

    def get_bitfield(self, address: bytes, page: int):
        with self.lock:
            p = PaginatedBitfield(False, self._state._db)
            return p.get_paginated_data(address, page)

    def is_slave(self, master_address: bytes, slave_pk: bytes) -> bool:
        with self.lock:
            slave_meta_data = self._state.get_slave_pk_access_type(master_address, slave_pk)
            if slave_meta_data:
                return slave_meta_data.access_type == 0
            return False

    def get_slave_pk_access_type(self, address: bytes, slave_pk: bytes) -> qbit_pb2.SlaveMetadata:
        with self.lock:
            return self._state.get_slave_pk_access_type(address, slave_pk)

    def get_transaction_hashes(self, address: bytes, item_index: int) -> list:
        p = PaginatedData(b'p_tx_hash', False, self._state._db)
        with self.lock:
            return p.get_paginated_data(address, item_index)

    def get_multi_sig_spend_txn_hashes(self, multi_sig_address: bytes, item_index: int) -> list:
        p = PaginatedData(b'p_multi_sig_spend', False, self._state._db)
        with self.lock:
            return p.get_paginated_data(multi_sig_address, item_index)

    def get_token_transaction_hashes(self, address: bytes, item_index: int) -> list:
        p = PaginatedData(b'p_tokens', False, self._state._db)
        with self.lock:
            return p.get_paginated_data(address, item_index)

    def get_slave_transaction_hashes(self, address: bytes, item_index: int) -> list:
        p = PaginatedData(b'p_slaves', False, self._state._db)
        with self.lock:
            return p.get_paginated_data(address, item_index)

    def get_lattice_pks_transaction_hashes(self, address: bytes, item_index: int) -> list:
        p = PaginatedData(b'p_lattice_pk', False, self._state._db)
        with self.lock:
            return p.get_paginated_data(address, item_index)

    def get_multi_sig_addresses(self, address: bytes, item_index: int) -> list:
        p = PaginatedData(b'p_multisig_address', False, self._state._db)
        with self.lock:
            return p.get_paginated_data(address, item_index)

    def get_inbox_message_transaction_hashes(self, address: bytes, item_index: int) -> list:
        p = PaginatedData(b'p_inbox_message', False, self._state._db)
        with self.lock:
            return p.get_paginated_data(address, item_index)

    def get_vote_stats(self, multi_sig_spend_txn_hash: bytes) -> VoteStats:
        with self.lock:
            return VoteStats.get_state(state=self._state, shared_key=multi_sig_spend_txn_hash)

    def get_token(self, address: bytes, token_txhash: bytes) -> qbit_pb2.TokenBalance:
        with self.lock:
            return self._state.get_token(address, token_txhash)

    def validate_all(self, tx: Transaction, check_nonce: bool) -> bool:
        with self.lock:
            addresses_set = set()
            tx.set_affected_address(addresses_set)
            state_container = self.new_state_container(addresses_set, self.height, True, None)
            if state_container is None:
                return False
            if not self.update_state_container(tx, state_container):
                return False
            return tx.validate_all(state_container, check_nonce)

    def validate_tx(self,
                    tx: Transaction,
                    addresses_state: dict,
                    addresses_bitfield: dict,
                    tokens: dict,
                    slaves: dict) -> bool:
        with self.lock:
            return self._state.validate_tx(tx, addresses_state, addresses_bitfield, tokens, slaves)

    def apply_txn(self, tx: Transaction, state_container: StateContainer) -> bool:
        with self.lock:
            return tx.apply(self._state, state_container)

    def get_tx_metadata(self, transaction_hash) -> list:
        with self.lock:
            return TransactionMetadata.get_tx_metadata(self._state, transaction_hash)

    def get_last_transactions(self):
        with self.lock:
            return LastTransactions.get_last_txs(self._state)

    def get_unconfirmed_transaction(self, transaction_hash) -> list:
        with self.lock:
            for tx_set in self.tx_pool.transactions:
                tx = tx_set[1].transaction
                if tx.txhash == transaction_hash:
                    return [tx, tx_set[1].timestamp]
            if transaction_hash in self.tx_pool.pending_tx_pool_hash:
                for tx_set in self.tx_pool.pending_tx_pool:
                    tx = tx_set[1].transaction
                    if tx.txhash == transaction_hash:
                        return [tx, tx_set[1].timestamp]

            return []

    def get_block_metadata(self, header_hash: bytes) -> Optional[BlockMetadata]:
        with self.lock:
            return BlockMetadata.get_block_metadata(self._state, header_hash)

    def get_blockheader_and_metadata(self, block_number=None) -> Tuple:
        with self.lock:
            # If block_number is None, use current height. If it's 0 or positive, use that exact number
            if block_number is None:
                block_number = self.height

            result = (None, None)
            block = self.get_block_by_number(block_number)
            if block:
                blockheader = block.blockheader
                blockmetadata = self.get_block_metadata(blockheader.headerhash)
                result = (blockheader, blockmetadata)

            return result

    def get_block_to_mine(self, miner, wallet_address) -> list:
        with miner.lock:  # Trying to acquire miner.lock to make sure pre_block_logic is not running
            with self.lock:
                last_block = self.last_block
                last_block_metadata = self.get_block_metadata(last_block.headerhash)
                return miner.get_block_to_mine(wallet_address,
                                               self.tx_pool,
                                               last_block,
                                               last_block_metadata.block_difficulty)

    def get_block_size_limit(self, block: Block, dev_config: DevConfig):
        with self.lock:
            return Block.get_block_size_limit(self._state, block, dev_config)

    def get_block_is_duplicate(self, block: Block) -> bool:
        with self.lock:
            return block.get_block(self._state, block.headerhash) is not None

    def get_config_by_block_number(self, block_number: int) -> config.DevConfig:
        dev_config = config.dev
        if block_number == 0:
            return dev_config

        with self.lock:
            while dev_config.activation_block_number >= block_number:
                dev_config_pb_data = self._state.get_dev_config_state(dev_config.prev_state_key)
                dev_config = config.DevConfig(dev_config_pb_data, True, True)
            return dev_config

    def get_seed_block(self, blockheader: BlockHeader):
        qn = Qryptonight()
        seed_height = qn.get_seed_height(blockheader.block_number)

        # If parent block belongs to main chain, then seed block will also be in the main chain
        prev_mainchain_block = self.get_block_by_number(blockheader.block_number - 1)
        if prev_mainchain_block.headerhash == blockheader.prev_headerhash:
            return self.get_block_by_number(seed_height)

        prev_block = self.get_block(blockheader.prev_headerhash)
        while prev_block.block_number > seed_height:
            prev_mainchain_block = self.get_block_by_number(prev_block.block_number)
            if prev_mainchain_block.headerhash == prev_block.headerhash:
                return self.get_block_by_number(seed_height)
            prev_block = self.get_block(prev_block.prev_headerhash)

        return prev_block

    def validate_mining_nonce(self, blockheader: BlockHeader, dev_config: config.DevConfig, enable_logging=True):
        with self.lock:
            parent_metadata = BlockMetadata.get_block_metadata(self._state, blockheader.prev_headerhash)
            parent_block = Block.get_block(self._state, blockheader.prev_headerhash)

            measurement = self.get_measurement(dev_config,
                                               blockheader.timestamp,
                                               blockheader.prev_headerhash,
                                               parent_metadata)
            diff, target = DifficultyTracker.get(
                measurement=measurement,
                parent_difficulty=parent_metadata.block_difficulty,
                dev_config=dev_config)

            mining_blob = blockheader.mining_blob(dev_config)
            qn = Qryptonight()
            seed_block = self.get_seed_block(blockheader)

            if enable_logging:
                logger.debug('-----------------START--------------------')
                logger.debug('Validate                #%s', blockheader.block_number)
                logger.debug('block.timestamp         %s', blockheader.timestamp)
                logger.debug('parent_block.timestamp  %s', parent_block.timestamp)
                logger.debug('parent_block.difficulty %s', UInt256ToString(parent_metadata.block_difficulty))
                logger.debug('diff                    %s', UInt256ToString(diff))
                logger.debug('target                  %s', bin2hstr(target))
                logger.debug('mining blob             %s', bin2hstr(mining_blob))
                logger.debug('seed_block              #%s', seed_block.block_number)
                logger.debug('seed_block hash         %s', bin2hstr(seed_block.headerhash))
                logger.debug('-------------------END--------------------')

            if not PoWValidator().verify_input(blockheader.block_number,
                                               seed_block.block_number,
                                               seed_block.headerhash,
                                               mining_blob,
                                               target):
                if enable_logging:
                    logger.warning("PoW verification failed")
                    tmp_hash = qn.hash(blockheader.block_number,
                                       seed_block.block_number,
                                       seed_block.headerhash,
                                       blockheader.mining_blob(dev_config))
                    logger.warning("{}".format(bin2hstr(tmp_hash)))
                    logger.debug('%s', blockheader.to_json())
                return False

            return True

    def get_headerhashes(self, start_blocknumber):
        with self.lock:
            start_blocknumber = max(0, start_blocknumber)
            end_blocknumber = min(self._last_block.block_number,
                                  start_blocknumber + 2 * config.dev.reorg_limit)

            total_expected_headerhash = end_blocknumber - start_blocknumber + 1

            node_header_hash = qbit_pb2.NodeHeaderHash()
            node_header_hash.block_number = start_blocknumber

            block = Block.get_block_by_number(self._state, end_blocknumber)
            block_headerhash = block.headerhash
            node_header_hash.headerhashes.append(block_headerhash)
            end_blocknumber -= 1

            while end_blocknumber >= start_blocknumber:
                block_metadata = BlockMetadata.get_block_metadata(self._state, block_headerhash)
                for headerhash in block_metadata.last_N_headerhashes[-1::-1]:
                    node_header_hash.headerhashes.append(headerhash)
                end_blocknumber -= len(block_metadata.last_N_headerhashes)
                if len(block_metadata.last_N_headerhashes) == 0:
                    break
                block_headerhash = block_metadata.last_N_headerhashes[0]

            node_header_hash.headerhashes[:] = node_header_hash.headerhashes[-1::-1]
            del node_header_hash.headerhashes[:len(node_header_hash.headerhashes) - total_expected_headerhash]

            return node_header_hash

    def set_broadcast_tx(self, broadcast_tx):
        with self.lock:
            self.tx_pool.set_broadcast_tx(broadcast_tx)

    def load(self, genesis_block):
        # load() has the following tasks:
        # Write Genesis Block into State immediately
        # Register block_number <-> blockhash mapping
        # Calculate difficulty Metadata for Genesis Block
        # Generate AddressStates from Genesis Block balances
        # Apply Genesis Block's transactions to the state
        # Detect if we are forked from genesis block and if so initiate recovery.

        # Loading Current Dev Config from State
        current_state_key = self._state.get_dev_config_current_state_key()
        if current_state_key is None:
            dev_config = config.DevConfig.create(None,
                                                 config.DevConfig.get_state_key(genesis_block.headerhash),
                                                 genesis_block.headerhash,
                                                 0,
                                                 ignore_check=True)
            dev_config_pbdata = dev_config.pbdata
            self._state.put_dev_config_state(dev_config_pbdata, None)
            current_state_key = dev_config.current_state_key
            self._state.put_dev_config_current_state_key(current_state_key, None)

        dev_config_pbdata = self._state.get_dev_config_state(current_state_key)
        config.dev.update_from_pbdata(dev_config_pbdata)

        state_migration = StateMigration()
        is_state_migration_needed = self._state.is_older_state_version()
        state_version = self._state.get_state_version()
        if is_state_migration_needed:
            if state_version == 0:
                state_migration.state_migration_step_1(self._state)

        height = self._state.get_mainchain_height()

        if height == -1:
            Block.put_block(self._state, genesis_block, None)
            block_number_mapping = qbit_pb2.BlockNumberMapping(headerhash=genesis_block.headerhash,
                                                              prev_headerhash=genesis_block.prev_headerhash)
            Block.put_block_number_mapping(self._state, genesis_block.block_number, block_number_mapping, None)
            parent_difficulty = StringToUInt256(str(config.user.genesis_difficulty))

            self.current_difficulty, _ = DifficultyTracker.get(
                measurement=config.dev.block_timing_in_seconds,
                parent_difficulty=parent_difficulty,
                dev_config=config.dev)

            block_metadata = BlockMetadata.create()
            block_metadata.set_block_difficulty(self.current_difficulty)
            block_metadata.set_cumulative_difficulty(self.current_difficulty)

            BlockMetadata.put_block_metadata(self._state, genesis_block.headerhash, block_metadata, None)
            address_set = set()

            coinbase_tx = Transaction.from_pbdata(genesis_block.transactions[0])
            coinbase_tx.set_affected_address(address_set)
            state_container = self.new_state_container(address_set,
                                                       0,
                                                       True,
                                                       None)
            
            # Initialize coinbase address with total remaining supply after genesis allocation
            total_genesis_allocation = 0
            for genesis_balance in GenesisBlock().genesis_balance:
                bytes_addr = genesis_balance.address
                state_container.addresses_state[bytes_addr] = OptimizedAddressState.get_default(bytes_addr)
                state_container.addresses_state[bytes_addr]._data.balance = genesis_balance.balance
                total_genesis_allocation += genesis_balance.balance
                
            # Calculate remaining supply for coinbase (for mining rewards)
            total_supply = int(config.dev.max_coin_supply * config.dev.quark_per_qbitcoin)
            coinbase_initial_balance = total_supply - total_genesis_allocation
            
            # Initialize coinbase address
            state_container.addresses_state[config.dev.coinbase_address] = OptimizedAddressState.get_default(config.dev.coinbase_address)
            state_container.addresses_state[config.dev.coinbase_address]._data.balance = coinbase_initial_balance

            for tx_idx in range(1, len(genesis_block.transactions)):
                tx = Transaction.from_pbdata(genesis_block.transactions[tx_idx])
                for addr in tx.addrs_to:
                    state_container.addresses_state[addr] = OptimizedAddressState.get_default(addr)

            if not isinstance(coinbase_tx, CoinBase):
                return False

            state_container.addresses_state[coinbase_tx.addr_to] = OptimizedAddressState.get_default(coinbase_tx.addr_to)

            if not coinbase_tx.validate_all(state_container):
                return False

            coinbase_tx.apply(self._state, state_container)

            for tx_idx in range(1, len(genesis_block.transactions)):
                tx = Transaction.from_pbdata(genesis_block.transactions[tx_idx])
                tx.apply(self._state, state_container)

            state_container.paginated_tx_hash.put_paginated_data(None)

            AddressState.put_addresses_state(self._state, state_container.addresses_state)
            state_container.paginated_bitfield.put_addresses_bitfield(None)
            TransactionMetadata.update_tx_metadata(self._state, genesis_block, None)
            self._state.update_mainchain_height(0, None)
        else:
            self._last_block = self.get_block_by_number(height)
            self.current_difficulty = BlockMetadata.get_block_metadata(self._state,
                                                                       self._last_block.headerhash).block_difficulty
            fork_state = self._state.get_fork_state()
            if fork_state:
                block = Block.get_block(self._state, fork_state.initiator_headerhash)
                self._fork_recovery(block, fork_state)

        if is_state_migration_needed:
            if state_version == 0:
                logger.warning("Please Wait... Starting State Migration From Version 0 to %s",
                               self._state.state_version)
                height = state_migration.height_from_state_version_0()
                start_block_number = self._state.get_mainchain_height() + 1
                logger.warning("Start blockheight %s", start_block_number)
                for block_number in range(start_block_number, height + 1):
                    block = state_migration.block_from_state_version_0(block_number)
                    if not self.add_block(block, check_stale=False):
                        print("System Exitting, due to migration failure")
                        sys.exit(1)
                    if block_number % 1000 == 0:
                        logger.warning("Migrated Block %s/%s", block_number, height)

                if self.height % 1000 != 0:
                    logger.warning("Migrated Block %s/%s", self.height, height)
                state_migration.state_migration_step_2(self._state)

            if state_version < 2:
                logger.warning("Please Wait... Starting State Migration From Version 1 to %s", self._state.state_version)
                height = self._state.get_mainchain_height()
                start_block_number = 1
                logger.warning("Start blockheight %s", start_block_number)
                total_block_reward = 0
                for block_number in range(start_block_number, height + 1):
                    total_block_reward += int(block_reward(block_number, config.dev))
                    if block_number % 1000 == 0:
                        logger.warning("Migrated Block %s/%s", block_number, height)

                if self.height % 1000 != 0:
                    logger.warning("Migrated Block %s/%s", self.height, height)

                logger.warning('Please Wait... While verifying State')
                total_balance = 0
                count = 0
                for address, _ in self._state._db.db:
                    address_state = None
                    if AddressState.address_is_valid(address):
                        address_state = OptimizedAddressState.get_optimized_address_state(self._state, address)
                    elif MultiSigAddressState.address_is_valid(address):
                        address_state = self.get_multi_sig_address_state(address)

                    if not address_state:
                        continue

                    count += 1
                    total_balance += address_state.balance

                    if count % 1000 == 0:
                        logger.warning("Processed Address %s", count)

                if count % 1000 != 0:
                    logger.warning("Processed Address %s", count)

                coinbase_balance = int(config.dev.coin_remaining_at_genesis * config.dev.shor_per_quanta - total_block_reward)
                total_supply = total_balance + coinbase_balance
                if total_supply != config.dev.max_coin_supply * config.dev.shor_per_quanta:
                    logger.warning('Total Supply: %s', total_supply)
                    logger.warning('Total Max Coin Supply: %s', config.dev.max_coin_supply * config.dev.shor_per_quanta)
                    raise Exception('Total supply mismatch, State Verification failed')

                a = OptimizedAddressState.get_optimized_address_state(self._state, config.dev.coinbase_address)
                a.pbdata.balance = coinbase_balance
                addresses_state = {config.dev.coinbase_address: a}
                a.put_optimized_addresses_state(self._state, addresses_state)

                a = OptimizedAddressState.get_optimized_address_state(self._state, config.dev.coinbase_address)
                if a.balance != coinbase_balance:
                    raise Exception('Unexpected Coinbase balance')
                self._state.put_state_version()
            if state_version < 3:
                # Adding extra block reward lost in block #2078158
                coinbase_addr = OptimizedAddressState.get_optimized_address_state(self._state, config.dev.coinbase_address)
                coinbase_addr.pbdata.balance += int(block_reward(2078158, config.dev))
                addresses_state = {config.dev.coinbase_address: coinbase_addr}
                coinbase_addr.put_optimized_addresses_state(self._state, addresses_state)

                self._state.put_state_version()

    def _update_chainstate(self, block: Block, batch):
        self._last_block = block
        self._update_block_number_mapping(block, batch)
        self.tx_pool.remove_tx_in_block_from_pool(block)
        self._state.update_mainchain_height(block.block_number, batch)
        self._state.update_re_org_limit(block.block_number, batch)
        TransactionMetadata.update_tx_metadata(self._state, block, batch)
        if block.block_number >= config.dev.hard_fork_heights[2]:
            banned_addr = OptimizedAddressState.get_optimized_address_state(self._state, config.dev.banned_address[0])
            if banned_addr.pbdata.balance > 0:
                banned_addr.pbdata.balance = 0
                addresses_state = {banned_addr.address: banned_addr}
                banned_addr.put_optimized_addresses_state(self._state, addresses_state)

    def _try_branch_add_block(self, block, dev_config: DevConfig, check_stale=True) -> bool:
        """
        This function returns list of bool types. The first bool represent
        if the block has been added successfully and the second bool
        represent the fork_flag, which becomes true when a block triggered
        into fork recovery.
        :param block:
        :param batch:
        :return: [Added successfully, fork_flag]
        """
        batch = self._state.batch

        if self._last_block.headerhash == block.prev_headerhash:
            if not self._apply_state_changes(block, batch):
                return False

        Block.put_block(self._state, block, batch)

        last_block_metadata = BlockMetadata.get_block_metadata(self._state, self._last_block.headerhash)
        if last_block_metadata is None:
            logger.warning("Could not find log metadata for %s", bin2hstr(self._last_block.headerhash))
            return False

        last_block_difficulty = int(UInt256ToString(last_block_metadata.cumulative_difficulty))

        new_block_metadata = self._add_block_metadata(block, dev_config, batch)
        new_block_difficulty = int(UInt256ToString(new_block_metadata.cumulative_difficulty))

        if new_block_difficulty > last_block_difficulty:
            if self._last_block.headerhash != block.prev_headerhash:
                fork_state = qbitstateinfo_pb2.ForkState(initiator_headerhash=block.headerhash)
                self._state.put_fork_state(fork_state, batch)
                self._state.write_batch(batch)
                return self._fork_recovery(block, fork_state)

            self._update_chainstate(block, batch)
            if check_stale:
                self.tx_pool.check_stale_txn(self.new_state_container,
                                             self.update_state_container,
                                             block.block_number)
            self.trigger_miner = True

        self._state.write_batch(batch)

        return True

    def _remove_block_from_mainchain(self, block: Block, latest_block_number: int, batch) -> bool:
        # Reverting Dev Config to older state, if any new config
        # was activated after addition of this block number
        if config.dev.activation_block_number == block.block_number:
            older_dev_config_pb_data = self._state.get_dev_config_state(config.dev.prev_state_key)
            self._state.put_dev_config_current_state_key(config.dev.prev_state_key, batch)
            config.dev.update_from_pbdata(older_dev_config_pb_data)

        if not self.revert_state_changes(block, batch):
            logger.warning("Fork Recovery: Revert State Changes Failed")
            return False

        self.tx_pool.add_tx_from_block_to_pool(block, latest_block_number)
        self._state.update_mainchain_height(block.block_number - 1, batch)
        TransactionMetadata.rollback_tx_metadata(self._state, block, batch)
        Block.remove_blocknumber_mapping(self._state, block.block_number, batch)

        return True

    def _get_fork_point(self, block: Block):
        tmp_block = block
        hash_path = []
        while True:
            if not block:
                raise Exception('[get_state] No Block Found %s, Initiator %s', block.headerhash, tmp_block.headerhash)

            mainchain_block = Block.get_block_by_number(self._state, block.block_number)
            if mainchain_block and mainchain_block.headerhash == block.headerhash:
                break

            if block.block_number == 0:
                raise Exception('[get_state] Alternate chain genesis is different, Initiator %s', tmp_block.headerhash)
            hash_path.append(block.headerhash)
            block = Block.get_block(self._state, block.prev_headerhash)

        return block.headerhash, hash_path

    def _rollback(self, forked_header_hash: bytes, fork_state: qbitstateinfo_pb2.ForkState = None):
        """
        Rollback from last block to the block just before the forked_header_hash
        :param forked_header_hash:
        :param fork_state:
        :return:
        """
        hash_path = []
        while self._last_block.headerhash != forked_header_hash:
            block = Block.get_block(self._state, self._last_block.headerhash)
            mainchain_block = Block.get_block_by_number(self._state, block.block_number)

            if block is None:
                logger.warning("self.state.get_block(self.last_block.headerhash) returned None")

            if mainchain_block is None:
                logger.warning("self.get_block_by_number(block.block_number) returned None")

            if block.headerhash != mainchain_block.headerhash:
                break
            hash_path.append(self._last_block.headerhash)

            batch = self._state.batch
            if not self._remove_block_from_mainchain(self._last_block, block.block_number, batch):
                return hash_path, False

            if fork_state:
                fork_state.old_mainchain_hash_path.extend([self._last_block.headerhash])
                self._state.put_fork_state(fork_state, batch)

            self._state.write_batch(batch)

            self._last_block = Block.get_block(self._state, self._last_block.prev_headerhash)

        return hash_path, True

    def add_chain(self, hash_path: list, fork_state: qbitstateinfo_pb2.ForkState) -> bool:
        """
        Add series of blocks whose headerhash mentioned into hash_path
        :param hash_path:
        :param fork_state:
        :param batch:
        :return:
        """
        with self.lock:
            start = 0
            try:
                start = hash_path.index(self._last_block.headerhash) + 1
            except ValueError:
                # Following condition can only be true if the fork recovery was interrupted last time
                if self._last_block.headerhash in fork_state.old_mainchain_hash_path:
                    return False

            for i in range(start, len(hash_path)):
                header_hash = hash_path[i]
                block = Block.get_block(self._state, header_hash)

                batch = self._state.batch

                if not self._apply_state_changes(block, batch):
                    return False

                self._update_chainstate(block, batch)

                logger.debug('Apply block #%d - [batch %d | %s]', block.block_number, i, hash_path[i])
                self._state.write_batch(batch)

            self._state.delete_fork_state()

            return True

    def _fork_recovery(self, block: Block, fork_state: qbitstateinfo_pb2.ForkState) -> bool:
        logger.info("Triggered Fork Recovery")
        # This condition only becomes true, when fork recovery was interrupted
        if fork_state.fork_point_headerhash:
            logger.info("Recovering from last fork recovery interruption")
            forked_header_hash, hash_path = fork_state.fork_point_headerhash, fork_state.new_mainchain_hash_path
        else:
            forked_header_hash, hash_path = self._get_fork_point(block)
            fork_state.fork_point_headerhash = forked_header_hash
            fork_state.new_mainchain_hash_path.extend(hash_path)
            self._state.put_fork_state(fork_state)

        rollback_done = False
        if fork_state.old_mainchain_hash_path:
            b = Block.get_block(self._state, fork_state.old_mainchain_hash_path[-1])
            if b and b.prev_headerhash == fork_state.fork_point_headerhash:
                rollback_done = True

        success = True
        if not rollback_done:
            logger.info("Rolling back")
            old_hash_path, success = self._rollback(forked_header_hash, fork_state)
        else:
            old_hash_path = fork_state.old_mainchain_hash_path

        if not success or not self.add_chain(hash_path[-1::-1], fork_state):
            logger.warning("Fork Recovery Failed... Recovering back to old mainchain")
            # If above condition is true, then it means, the node failed to add_chain
            # Thus old chain state, must be retrieved
            self._rollback(forked_header_hash)
            self.add_chain(old_hash_path[-1::-1], fork_state)  # Restores the old chain state
            return False

        logger.info("Fork Recovery Finished")

        self.trigger_miner = True
        return True

    def _add_block(self, block, check_stale=True) -> bool:
        dev_config = self.get_config_by_block_number(block.block_number)
        self.trigger_miner = False

        block_size_limit = self.get_block_size_limit(block, dev_config)
        if block_size_limit and block.size > block_size_limit:
            logger.info('Block Size greater than threshold limit %s > %s', block.size, block_size_limit)
            return False

        return self._try_branch_add_block(block, dev_config, check_stale)

    def add_block(self, block: Block, check_stale=True) -> bool:
        with self.lock:
            if block.block_number <= self.re_org_limit:
                logger.debug('Skipping block #%s as beyond re-org limit', block.block_number)
                return False

            if self.get_block_is_duplicate(block):
                logger.warning("Duplicate Block found #%s", block.block_number)
                return False

            block_flag = self._add_block(block, check_stale=check_stale)
            if not block_flag:
                logger.warning("[ChainManager] Failed to Add Block #%s", block.block_number)
                return False

            logger.info('Added Block #%s %s', block.block_number, bin2hstr(block.headerhash))
            return True

    def _add_block_metadata(self,
                            block: Block,
                            dev_config: DevConfig,
                            batch):
        block_metadata = BlockMetadata.get_block_metadata(self._state, block.headerhash)
        if not block_metadata:
            block_metadata = BlockMetadata.create()

        parent_metadata = BlockMetadata.get_block_metadata(self._state, block.prev_headerhash)

        parent_block_difficulty = parent_metadata.block_difficulty
        parent_cumulative_difficulty = parent_metadata.cumulative_difficulty

        block_metadata.update_last_headerhashes(parent_metadata.last_N_headerhashes, block.prev_headerhash)
        measurement = self.get_measurement(dev_config, block.timestamp, block.prev_headerhash, parent_metadata)

        block_difficulty, _ = DifficultyTracker.get(
            measurement=measurement,
            parent_difficulty=parent_block_difficulty,
            dev_config=dev_config)

        block_cumulative_difficulty = StringToUInt256(str(
            int(UInt256ToString(block_difficulty)) +
            int(UInt256ToString(parent_cumulative_difficulty))))

        block_metadata.set_block_difficulty(block_difficulty)
        block_metadata.set_cumulative_difficulty(block_cumulative_difficulty)

        parent_metadata.add_child_headerhash(block.headerhash)
        BlockMetadata.put_block_metadata(self._state, block.prev_headerhash, parent_metadata, batch)
        BlockMetadata.put_block_metadata(self._state, block.headerhash, block_metadata, batch)

        return block_metadata

    def _update_block_number_mapping(self, block, batch):
        block_number_mapping = qbit_pb2.BlockNumberMapping(headerhash=block.headerhash,
                                                          prev_headerhash=block.prev_headerhash)
        Block.put_block_number_mapping(self._state, block.block_number, block_number_mapping, batch)

    @staticmethod
    def set_affected_address(block: Block) -> set:
        addresses_set = set()
        for proto_tx in block.transactions:
            tx = Transaction.from_pbdata(proto_tx)
            tx.set_affected_address(addresses_set)

        for genesis_balance in GenesisBlock().genesis_balance:
            bytes_addr = genesis_balance.address
            if bytes_addr not in addresses_set:
                addresses_set.add(bytes_addr)

        return addresses_set

    def new_state_container(self,
                            address_set: set,
                            block_number: int,
                            write_access: bool,
                            batch) -> Optional[StateContainer]:
        tokens = Indexer(b'token', self._state._db)
        slaves = Indexer(b'slave', self._state._db)
        lattice_pk = Indexer(b'lattice_pk', self._state._db)
        multi_sig_spend_txs = dict()
        votes_stats = dict()

        dev_config = self.get_config_by_block_number(block_number)

        addresses_state, success = self.get_state_mainchain(address_set)
        if not success:
            logger.warning("Failed to get state mainchain for coinbase_tx")
            return None

        return StateContainer(addresses_state,
                              tokens,
                              slaves,
                              lattice_pk,
                              multi_sig_spend_txs,
                              votes_stats,
                              block_number,
                              self._state.total_coin_supply,
                              dev_config,
                              write_access,
                              self._state._db,
                              batch)

    def update_state_container(self,
                               tx: Transaction,
                               state_container: StateContainer) -> bool:
        address_set = set()
        tx.set_affected_address(address_set)
        tokens = Indexer(b'token', self._state._db)
        slaves = Indexer(b'slave', self._state._db)
        lattice_pk = Indexer(b'lattice_pk', self._state._db)
        multi_sig_spend_txs = dict()
        votes_stats = dict()

        if not isinstance(tx, CoinBase):
            slave_addr = Transaction.get_slave(tx)
            if slave_addr is not None:
                key = (tx.addr_from, tx.PK)
                if key not in state_container.slaves.data:
                    slaves.load(key, qbit_pb2.SlaveMetadata())

        if isinstance(tx, TransferTokenTransaction):
            key = (tx.addr_from, tx.token_txhash)
            if key not in state_container.tokens.data:
                tokens.load(key, qbit_pb2.TokenBalance())

            for address in tx.addrs_to:
                key = (address, tx.token_txhash)
                if key in state_container.tokens.data:
                    continue
                tokens.load(key, qbit_pb2.TokenBalance())
        elif isinstance(tx, SlaveTransaction):
            for slave_pk in tx.slave_pks:
                key = (tx.addr_from, slave_pk)
                if key not in state_container.slaves.data:
                    slaves.load(key, qbit_pb2.SlaveMetadata())
        elif isinstance(tx, LatticeTransaction):
            key = (tx.addr_from, tx.pk1, tx.pk2, tx.pk3)
            if key not in state_container.lattice_pk.data:
                lattice_pk.load(key, qbit_pb2.LatticePKMetadata())
        # elif isinstance(tx, TokenTransaction):
        #     for initial_balance in tx.initial_balances:
        #         if (initial_balance.address, tx.txhash) not in tokens:
        #             tokens[(initial_balance.address, tx.txhash)] = 0
        elif isinstance(tx, MultiSigVote):
            if tx.shared_key not in state_container.multi_sig_spend_txs:
                tx_meta_data = TransactionMetadata.get_tx_metadata(self._state, tx.shared_key)
                if tx_meta_data is None:
                    logger.warning("[MultiSigVote] TransactionMetaData not found for the shared_key %s", tx.shared_key)
                    return False

                multi_sig_spend_tx = tx_meta_data[0]
                multi_sig_spend_txs[tx.shared_key] = multi_sig_spend_tx

                if tx.shared_key not in state_container.votes_stats:
                    votes_stats[tx.shared_key] = VoteStats.get_state(self._state,
                                                                     tx.shared_key)

                # Adding address whose address state will be affected on the execution of multi_sig txn.
                if multi_sig_spend_tx.multi_sig_address not in state_container.addresses_state:
                    address_set.add(multi_sig_spend_tx.multi_sig_address)
                for address in multi_sig_spend_tx.addrs_to:
                    if address not in state_container.addresses_state:
                        address_set.add(address)

        addresses_state, success = self.get_state_mainchain(address_set,
                                                            ignore_addresses_set=state_container.addresses_state.keys())
        if not success:
            logger.warning("get_state_mainchain failed")
            return False

        return state_container.update(addresses_state,
                                      tokens,
                                      slaves,
                                      lattice_pk,
                                      multi_sig_spend_txs,
                                      votes_stats)

    def _apply_state_changes(self, block, batch) -> bool:
        state_container = self.new_state_container(set(),
                                                   block.block_number,
                                                   True,
                                                   batch)
        if state_container is None:
            return False

        # Processing Rest of the Transaction
        for index, proto_tx in enumerate(block.transactions):
            if index > 0 and block.block_number != 2078158:
                if proto_tx.WhichOneof('transactionType') == 'coinbase':
                    logger.warning("Multiple coinbase transaction found")
                    return False

            tx = Transaction.from_pbdata(proto_tx)
            if not self.update_state_container(tx, state_container):
                return False

            if not tx.validate_all(state_container):
                return False

            tx.apply(self._state, state_container)

        # This should be done before put_addresses_state & paginated_tx_hash
        # as it make changes on the balance and add more txn hash that needed
        # to be added
        VoteStats.put_all(self._state, state_container)

        state_container.paginated_tx_hash.put_paginated_data(batch)

        state_container.paginated_lattice_pk.put_paginated_data(batch)

        state_container.paginated_multisig_address.put_paginated_data(batch)

        state_container.paginated_multi_sig_spend.put_paginated_data(batch)

        state_container.paginated_inbox_message.put_paginated_data(batch)

        # TODO: Add Key value storage to lattice pk
        state_container.tokens.put(batch)
        # This is needed to show list of tokens owned by an address on Web Wallet
        state_container.paginated_tokens_hash.put_paginated_data(batch)

        state_container.slaves.put(batch)
        # This is to retrieve the list of slaves of any particular address
        # Could be removed if not needed
        state_container.paginated_slaves_hash.put_paginated_data(batch)

        state_container.lattice_pk.put(batch)

        state_container.paginated_bitfield.put_addresses_bitfield(batch)

        AddressState.put_addresses_state(self._state, state_container.addresses_state, batch)

        return True

    def revert_state_changes(self, block, batch) -> bool:
        address_set = set()

        coinbase_tx = Transaction.from_pbdata(block.transactions[0])
        coinbase_tx.set_affected_address(address_set)
        state_container = self.new_state_container(address_set,
                                                   block.block_number,
                                                   True,
                                                   batch)
        if state_container is None:
            return False

        # Processing Rest of the Transaction
        len_transactions = len(block.transactions)
        for tx_idx in range(len_transactions - 1, 0, -1):
            tx = Transaction.from_pbdata(block.transactions[tx_idx])
            if not self.update_state_container(tx, state_container):
                logger.warning("Failed to update state_container in fork recovery")
                return False

            addr_from_pk_state = state_container.addresses_state[tx.addr_from]
            addr_from_pk = Transaction.get_slave(tx)
            if addr_from_pk:
                addr_from_pk_state = state_container.addresses_state[addr_from_pk]

            if tx.nonce != addr_from_pk_state.nonce:
                logger.warning('nonce incorrect while reverting state')
                logger.warning('subtype: %s', tx.type)
                logger.warning('%s actual: %s expected: %s', tx.addr_from, tx.nonce, addr_from_pk_state.nonce)
                return False

            if not state_container.paginated_bitfield.load_bitfield_and_ots_key_reuse(addr_from_pk_state.address, tx.ots_key):
                logger.warning('pubkey reuse not detected: invalid tx %s', bin2hstr(tx.txhash))
                logger.warning('subtype: %s', tx.type)
                return False

            if not tx.revert(self._state, state_container):
                return False

        # Reverting CoinBase Txn
        coinbase_tx.revert(self._state, state_container)

        # Invalid if slave_key doesnt exists in state
        for address_slave_pk in state_container.slaves.data:
            address, slave_pk = address_slave_pk
            access_type = self.get_slave_pk_access_type(address, slave_pk)
            if access_type is None:
                logger.warning("Failed No Access Type Found")
                return False

        # This should be done before put_addresses_state, paginated_tx_hash &
        # after vote_txn as it make changes on the balance and add more txn hash that
        # needed to be removed, as well as total weight is checked before
        # reverting the multi sig execution.
        if not VoteStats.revert_all(self._state, state_container):
            logger.warning("Failed to Revert Vote Stats")
            return False

        if not state_container.paginated_tx_hash.put_paginated_data(batch):
            logger.warning("Failed Revert Addresses Transaction Hashes")
            return False

        if not state_container.paginated_lattice_pk.put_paginated_data(batch):
            logger.warning("Failed Revert Addresses Lattice PK Hashes")
            return False

        # TODO: Put token should delete tokens address map with 0 balance
        state_container.tokens.put(batch)

        if not state_container.paginated_tokens_hash.put_paginated_data(batch):
            logger.warning("Failed Revert Tokens Hash")
            return False

        state_container.slaves.put(batch)

        if not state_container.paginated_slaves_hash.put_paginated_data(batch):
            logger.warning("Failed Revert Slaves Hash")
            return False

        state_container.lattice_pk.put(batch)

        if not state_container.paginated_multi_sig_spend.put_paginated_data(batch):
            logger.warning("Failed Revert Multi Sig Spend")
            return False

        if not state_container.paginated_inbox_message.put_paginated_data(batch):
            logger.warning("Failed Revert Inbox Message")
            return False

        if not state_container.paginated_multisig_address.put_paginated_data(batch):
            logger.warning("Failed Revert Multi Sig Addresses")
            return False

        state_container.paginated_bitfield.put_addresses_bitfield(batch)

        # All state which records page numbers, must be saved at last
        AddressState.put_addresses_state(self._state, state_container.addresses_state, batch)

        return True

    def get_state_mainchain(self, addresses_set: set, ignore_addresses_set: set = None) -> [dict, bool]:
        addresses_state = dict()
        for address in addresses_set:
            if ignore_addresses_set is not None:
                if address in ignore_addresses_set:
                    continue
            if OptimizedAddressState.address_is_valid(address) or address == config.dev.coinbase_address:
                addresses_state[address] = self.get_optimized_address_state(address)
            elif MultiSigAddressState.address_is_valid(address):
                multi_sig_address_state = MultiSigAddressState.get_multi_sig_address_state_by_address(self._state._db,
                                                                                                      address)
                addresses_state[address] = multi_sig_address_state

                # Load Address State of signatories as it needs to be processed by MultiSigSpend Txn
                # for inserting txn hash of MultiSigSpend to all signatories
                for signatory in multi_sig_address_state.signatories:
                    # Ignore signatore which are in ignore_addresses_set as their address state is already loaded
                    if ignore_addresses_set is not None:
                        if signatory in ignore_addresses_set:
                            continue
                    if OptimizedAddressState.address_is_valid(signatory):
                        addresses_state[signatory] = self.get_optimized_address_state(signatory)
                    elif MultiSigAddressState.address_is_valid(signatory):
                        addresses_state[signatory] = MultiSigAddressState.get_multi_sig_address_state_by_address(self._state._db,
                                                                                                                 signatory)
                    else:
                        return None, False
            else:
                return None, False

        return addresses_state, True

    def get_all_address_state(self) -> list:
        addresses_state = []

        try:
            for address, _ in self._state._db.db:
                if AddressState.address_is_valid(address) or address == config.dev.coinbase_address:
                    addresses_state.append(self.get_address_state(address).pbdata)
            return addresses_state
        except Exception as e:
            logger.error("Exception in get_all_address_state %s", e)

        return []

    def get_measurement(self,
                        dev_config: DevConfig,
                        block_timestamp,
                        parent_headerhash,
                        parent_metadata: BlockMetadata):
        count_headerhashes = len(parent_metadata.last_N_headerhashes)

        if count_headerhashes == 0:
            return dev_config.block_timing_in_seconds
        elif count_headerhashes == 1:
            nth_block = Block.get_block(self._state, parent_headerhash)
            count_headerhashes += 1
        else:
            nth_block = Block.get_block(self._state, parent_metadata.last_N_headerhashes[1])

        nth_block_timestamp = nth_block.timestamp
        if count_headerhashes < dev_config.N_measurement:
            nth_block_timestamp -= dev_config.block_timing_in_seconds

        return (block_timestamp - nth_block_timestamp) // count_headerhashes

    # TODO: This will be broken by stateful dev config
    @functools.lru_cache(maxsize=config.dev.block_timeseries_size + 50)
    def get_block_datapoint(self, headerhash):
        block = self.get_block(headerhash)
        if block is None:
            return None

        block_metadata = self.get_block_metadata(headerhash)
        prev_block_metadata = self.get_block_metadata(block.prev_headerhash)
        prev_block = self.get_block(block.prev_headerhash)

        data_point = qbit_pb2.BlockDataPoint()
        data_point.number = block.block_number
        data_point.header_hash = headerhash
        if prev_block is not None:
            data_point.header_hash_prev = prev_block.headerhash
        data_point.timestamp = block.timestamp
        data_point.time_last = 0
        data_point.time_movavg = 0
        data_point.difficulty = UInt256ToString(block_metadata.block_difficulty)

        if prev_block is not None:
            data_point.time_last = block.timestamp - prev_block.timestamp
            if prev_block.block_number == 0:
                data_point.time_last = config.dev.block_timing_in_seconds

            movavg = self.get_measurement(config.dev,
                                          block.timestamp,
                                          block.prev_headerhash,
                                          prev_block_metadata)
            data_point.time_movavg = movavg

            try:
                # FIXME: need to consider average difficulty here
                data_point.hash_power = int(data_point.difficulty) * (config.dev.block_timing_in_seconds / movavg)
            except ZeroDivisionError:
                data_point.hash_power = 0

        return data_point

    def get_unused_ots_index2(self, address, start_ots_index=0):
        # OTS tracking not needed for Falcon signatures
        return None

    def get_unused_ots_index(self,
                             addresses_bitfield: dict,
                             addresses_state: dict,
                             address,
                             paginated_bitfield: PaginatedBitfield,
                             start_ots_index=0):
        """
        Finds the unused ots index above the given start_ots_index.
        """
        if address not in addresses_state:
            addresses_state[address] = self.get_optimized_address_state(address)

        address_state = addresses_state[address]

        ots_key_count = (2 ** address_state.height)
        max_page = ceil(ots_key_count / config.dev.ots_tracking_per_page)
        if address_state.ots_bitfield_used_page == max_page:
            return None

        page = max(address_state.ots_bitfield_used_page, start_ots_index // config.dev.ots_tracking_per_page) + 1

        for i in range((page - 1) * config.dev.ots_tracking_per_page // 8, ots_key_count // 8):
            page = (i // config.dev.ots_tracking_per_page) + 1
            key = paginated_bitfield.generate_bitfield_key(address, page)
            if key not in addresses_bitfield:
                addresses_bitfield[key] = paginated_bitfield.get_paginated_data(address, page)

            ots_bitfield = addresses_bitfield[key]
            index = i % config.dev.ots_tracking_per_page
            if ots_bitfield[index][0] < 255:
                offset = 8 * index + (page - 1) * config.dev.ots_tracking_per_page
                bitfield = bytearray(ots_bitfield[index])
                for relative in range(0, 8):
                    if ((bitfield[0] >> relative) & 1) != 1:
                        if offset + relative >= start_ots_index:
                            return offset + relative

        return None



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.

import os
import json
import time
from typing import Optional
from cryptography.fernet import Fernet
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
import base64

from pyqrllib.pyqrllib import  hstr2bin

from qbitcoin.core.config import UserConfig
from qbitcoin.core.misc import logger
from qbitcoin.core.Singleton import Singleton


class DonorWalletManager(object, metaclass=Singleton):
    """
    Secure donor wallet management for staking rewards
    Handles encrypted storage of donor wallet private keys
    """
    
    def __init__(self):
        user_config = UserConfig.getInstance()
        self.donor_wallet_dir = os.path.join(user_config.qrl_dir, 'donor_wallets')
        self.donor_wallet_file = os.path.join(self.donor_wallet_dir, 'donor_wallet.json')
        self.loaded_wallet = None
        self.decrypted_private_key = None
        
        # Ensure donor wallet directory exists
        os.makedirs(self.donor_wallet_dir, exist_ok=True)
        
    def _derive_key_from_password(self, password: str, salt: bytes) -> bytes:
        """Derive encryption key from password using PBKDF2"""
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,
            salt=salt,
            iterations=100000,
        )
        key = base64.urlsafe_b64encode(kdf.derive(password.encode()))
        return key
        
    def _encrypt_data(self, data: str, password: str) -> dict:
        """Encrypt data with password"""
        salt = os.urandom(16)
        key = self._derive_key_from_password(password, salt)
        fernet = Fernet(key)
        encrypted_data = fernet.encrypt(data.encode())
        
        return {
            'encrypted_data': base64.b64encode(encrypted_data).decode(),
            'salt': base64.b64encode(salt).decode(),
            'type': 'encrypted_donor_wallet'
        }
        
    def _decrypt_data(self, encrypted_data: str, salt: str, password: str) -> str:
        """Decrypt data with password"""
        salt_bytes = base64.b64decode(salt.encode())
        key = self._derive_key_from_password(password, salt_bytes)
        fernet = Fernet(key)
        
        encrypted_bytes = base64.b64decode(encrypted_data.encode())
        decrypted_data = fernet.decrypt(encrypted_bytes)
        return decrypted_data.decode()
        
    def import_donor_wallet(self, wallet_file_path: str, password: str = None) -> bool:
        """
        Import an existing wallet file as donor wallet
        
        Args:
            wallet_file_path: Path to the wallet file (JSON format)
            password: Password if wallet is encrypted (None if not encrypted)
            
        Returns:
            bool: True if wallet was imported successfully
        """
        try:
            if not os.path.exists(wallet_file_path):
                logger.error("Wallet file not found: %s", wallet_file_path)
                return False
                
            # Read the wallet file
            with open(wallet_file_path, 'r') as f:
                if wallet_file_path.endswith('.json'):
                    # Unencrypted JSON wallet (like genesis_keys.json)
                    wallet_data = json.load(f)
                    
                    # Extract required fields
                    if 'address' not in wallet_data or 'private_key_hex' not in wallet_data:
                        logger.error("Invalid wallet format - missing required fields")
                        return False
                        
                    # Convert hex keys to bytes
                    private_key = bytes.fromhex(wallet_data['private_key_hex'])
                    public_key = bytes.fromhex(wallet_data.get('public_key_hex', ''))
                    
                    # Prepare standardized wallet data
                    standardized_data = {
                        'address': wallet_data['address'],
                        'private_key': private_key.hex(),
                        'public_key': public_key.hex(),
                        'imported_at': time.time(),
                        'algorithm': wallet_data.get('algorithm', 'falcon-512'),
                        'original_file': wallet_file_path
                    }
                    
                elif wallet_file_path.endswith('.enc'):
                    # Encrypted wallet file
                    if not password:
                        logger.error("Password required for encrypted wallet")
                        return False
                        
                    encrypted_data = f.read().encode() if isinstance(f.read(), str) else f.read()
                    decrypted_json = self._decrypt_data(encrypted_data, password)
                    standardized_data = json.loads(decrypted_json)
                    
                else:
                    logger.error("Unsupported wallet file format")
                    return False
            
            # Save as encrypted donor wallet
            donor_name = os.path.splitext(os.path.basename(wallet_file_path))[0]
            donor_password = password or "imported_wallet_default"
            
            encrypted_wallet_dict = self._encrypt_data(json.dumps(standardized_data), donor_password)
            
            # Save to donor wallet directory
            donor_wallet_path = os.path.join(self.donor_wallet_dir, f"{donor_name}_donor.json")
            with open(donor_wallet_path, 'w') as f:
                json.dump(encrypted_wallet_dict, f, indent=2)
                
            logger.info("Donor wallet imported successfully: %s", donor_wallet_path)
            
            # Store the password for this session
            self._current_password = donor_password
            
            return True
            
        except Exception as e:
            logger.error("Failed to import donor wallet: %s", str(e))
            return False
            
    def load_donor_wallet(self, password: str) -> Optional[dict]:
        """
        Load and decrypt donor wallet
        
        Args:
            password: Password for decryption
            
        Returns:
            dict: Wallet data if successful, None otherwise
        """
        try:
            # Look for any donor wallet files
            donor_wallet_files = []
            if os.path.exists(self.donor_wallet_file):
                donor_wallet_files.append(self.donor_wallet_file)
            
            # Also look for imported donor wallets
            if os.path.exists(self.donor_wallet_dir):
                for filename in os.listdir(self.donor_wallet_dir):
                    if filename.endswith('_donor.json'):
                        donor_wallet_files.append(os.path.join(self.donor_wallet_dir, filename))
            
            if not donor_wallet_files:
                logger.warning("No donor wallet files found")
                return None
            
            # Try to load the first available wallet
            for wallet_file in donor_wallet_files:
                try:
                    # Load encrypted data
                    with open(wallet_file, 'r') as f:
                        encrypted_wallet = json.load(f)
                        
                    # Verify it's encrypted wallet
                    if encrypted_wallet.get('type') != 'encrypted_donor_wallet':
                        logger.warning("Invalid wallet file format: %s", wallet_file)
                        continue
                        
                    # Decrypt wallet data
                    decrypted_data = self._decrypt_data(
                        encrypted_wallet['encrypted_data'],
                        encrypted_wallet['salt'],
                        password
                    )
                    
                    wallet_data = json.loads(decrypted_data)
                    
                    # Cache loaded wallet
                    self.loaded_wallet = wallet_data
                    self.decrypted_private_key = wallet_data['private_key']
                    
                    logger.info("Donor wallet loaded successfully: %s", wallet_data['address'])
                    return wallet_data
                    
                except Exception as e:
                    logger.warning("Failed to load wallet %s: %s", wallet_file, str(e))
                    continue
            
            logger.error("Could not load any donor wallet with provided password")
            return None
            
        except Exception as e:
            logger.error("Failed to load donor wallet: %s", str(e))
            return None
            
    def get_wallet_info(self) -> Optional[dict]:
        """Get current loaded wallet info (including public key but not private key)"""
        if not self.loaded_wallet:
            return None
            
        return {
            'address': self.loaded_wallet['address'],
            'public_key': self.loaded_wallet.get('public_key_hex') or self.loaded_wallet.get('public_key', ''),
            'algorithm': self.loaded_wallet.get('algorithm', 'falcon-512'),
            'imported_at': self.loaded_wallet.get('imported_at'),
            'created_at': self.loaded_wallet.get('created_at')
        }
        
    def get_address_bytes(self) -> Optional[bytes]:
        """Get donor wallet address as bytes"""
        if not self.loaded_wallet:
            return None
            
        try:
            address_hex = self.loaded_wallet['address'][1:]  # Remove Q prefix
            address_bytes = hstr2bin(address_hex)
            
            # Ensure it's bytes, not tuple
            if isinstance(address_bytes, tuple):
                return bytes(address_bytes)
            elif isinstance(address_bytes, bytes):
                return address_bytes
            else:
                return bytes(address_bytes)
        except Exception as e:
            logger.error("Failed to convert address to bytes: %s", str(e))
            return None
            
    def get_private_key_bytes(self) -> Optional[bytes]:
        """Get donor wallet private key as bytes"""
        if not self.decrypted_private_key:
            return None
            
        try:
            private_key_bytes = hstr2bin(self.decrypted_private_key)
            
            # Ensure it's bytes, not tuple
            if isinstance(private_key_bytes, tuple):
                return bytes(private_key_bytes)
            elif isinstance(private_key_bytes, bytes):
                return private_key_bytes
            else:
                return bytes(private_key_bytes)
        except Exception as e:
            logger.error("Failed to convert private key to bytes (security: details omitted)")
            return None
            
    def is_wallet_loaded(self) -> bool:
        """Check if a donor wallet is currently loaded"""
        return self.loaded_wallet is not None
        
    def unload_wallet(self):
        """Unload current wallet and clear sensitive data"""
        self.loaded_wallet = None
        self.decrypted_private_key = None
        logger.info("Donor wallet unloaded")
        
    def wallet_exists(self) -> bool:
        """Check if donor wallet file exists"""
        return os.path.exists(self.donor_wallet_file)
        
    def get_wallet_path(self) -> str:
        """Get the path to the donor wallet file"""
        return self.donor_wallet_file
        
    def verify_password(self, password: str) -> bool:
        """Verify if password is correct for the stored wallet"""
        try:
            if not self.wallet_exists():
                return False
                
            # Try to load wallet with password
            wallet_data = self.load_donor_wallet(password)
            return wallet_data is not None
            
        except Exception:
            return False
            
    def import_encrypted_wallet_as_unencrypted(self, wallet_path: str, password: str) -> bool:
        """Import encrypted wallet, decrypt it, and save as unencrypted for automatic loading"""
        try:
            # Load and decrypt the wallet
            with open(wallet_path, 'r') as f:
                encrypted_data = json.load(f)
                
            if not encrypted_data.get('encrypted'):
                logger.error('Wallet is not encrypted')
                return False
                
            # Decrypt the wallet data
            salt = base64.b64decode(encrypted_data['salt'])
            key = self._derive_key_from_password(password, salt)
            fernet = Fernet(key)
            
            try:
                decrypted_bytes = fernet.decrypt(encrypted_data['encrypted_data'].encode())
                wallet_data = json.loads(decrypted_bytes.decode())
            except Exception as e:
                logger.error('Failed to decrypt wallet (wrong password?): %s', str(e))
                return False
                
            # Save as unencrypted for automatic loading
            wallet_filename = os.path.basename(wallet_path)
            name_part = os.path.splitext(wallet_filename)[0]
            donor_wallet_path = os.path.join(self.donor_wallet_dir, f"{name_part}_donor.json")
            
            # Prepare unencrypted wallet data
            unencrypted_wallet = {
                'address': wallet_data['address'],
                'private_key_hex': wallet_data.get('private_key_hex') or wallet_data.get('pk'),
                'algorithm': wallet_data.get('algorithm', 'falcon-512'),
                'imported_from': wallet_path,
                'imported_at': time.time(),
                'encrypted': False  # Mark as unencrypted
            }
            
            # Save unencrypted
            with open(donor_wallet_path, 'w') as f:
                json.dump(unencrypted_wallet, f, indent=2)
                
            logger.info('Encrypted wallet decrypted and saved to: %s', donor_wallet_path)
            return True
            
        except Exception as e:
            logger.error('Error importing encrypted wallet: %s', str(e))
            return False
            
    def import_unencrypted_wallet_direct(self, wallet_path: str) -> bool:
        """Import unencrypted wallet directly without encryption for automatic loading"""
        try:
            # Load wallet data
            with open(wallet_path, 'r') as f:
                wallet_data = json.load(f)
                
            # Validate structure - support multiple wallet formats
            has_address = 'address' in wallet_data
            has_private_key = any(key in wallet_data for key in ['pk', 'private_key_hex', 'private_key'])
            
            if not (has_address and has_private_key):
                logger.error('Invalid wallet structure: missing address or private key')
                logger.error('Expected: address + (pk/private_key_hex/private_key)')
                logger.error('Found keys: %s', list(wallet_data.keys()))
                return False
                
            # Save as unencrypted for automatic loading
            wallet_filename = os.path.basename(wallet_path)
            name_part = os.path.splitext(wallet_filename)[0]
            donor_wallet_path = os.path.join(self.donor_wallet_dir, f"{name_part}_donor.json")
            
            # Extract private key from various possible field names
            private_key_hex = None
            if 'private_key_hex' in wallet_data:
                private_key_hex = wallet_data['private_key_hex']
            elif 'pk' in wallet_data:
                private_key_hex = wallet_data['pk']
            elif 'private_key' in wallet_data:
                private_key_hex = wallet_data['private_key']
            
            # Prepare unencrypted wallet data
            unencrypted_wallet = {
                'address': wallet_data['address'],
                'private_key_hex': private_key_hex,
                'algorithm': wallet_data.get('algorithm', 'falcon-512'),
                'imported_from': wallet_path,
                'imported_at': time.time(),
                'encrypted': False  # Mark as unencrypted
            }
            
            # Copy public key if available (important for transaction signing)
            if 'public_key_hex' in wallet_data:
                unencrypted_wallet['public_key_hex'] = wallet_data['public_key_hex']
            elif 'public_key' in wallet_data:
                unencrypted_wallet['public_key_hex'] = wallet_data['public_key']
            
            # Copy additional fields if present
            optional_fields = ['genesis_balance', 'genesis_balance_qrl', 'balance']
            for field in optional_fields:
                if field in wallet_data:
                    unencrypted_wallet[field] = wallet_data[field]
            
            # Save unencrypted
            with open(donor_wallet_path, 'w') as f:
                json.dump(unencrypted_wallet, f, indent=2)
                
            logger.info('Unencrypted wallet saved to: %s', donor_wallet_path)
            return True
            
        except Exception as e:
            logger.error('Error importing unencrypted wallet: %s', str(e))
            return False
            
    def load_unencrypted_donor_wallet(self, wallet_data: dict) -> bool:
        """Load donor wallet from unencrypted data"""
        try:
            # Validate wallet data
            if not isinstance(wallet_data, dict):
                logger.error('Invalid wallet data format')
                return False
                
            required_fields = ['address', 'private_key_hex']
            if not all(field in wallet_data for field in required_fields):
                logger.error('Missing required fields in wallet data')
                return False
                
            # Store wallet data
            self.loaded_wallet = {
                'address': wallet_data['address'],
                'algorithm': wallet_data.get('algorithm', 'falcon-512'),
                'public_key_hex': wallet_data.get('public_key_hex', ''),
                'imported_at': wallet_data.get('imported_at'),
                'created_at': wallet_data.get('created_at')
            }
            
            # Store decrypted private key
            self.decrypted_private_key = wallet_data['private_key_hex']
            
            logger.info('Unencrypted donor wallet loaded successfully')
            return True
            
        except Exception as e:
            logger.error('Error loading unencrypted donor wallet: %s', str(e))
            return False


# Import time for timestamp
import time



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: qbitlegacy.proto
# Protobuf Python Version: 6.30.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    30,
    0,
    '',
    'qbitlegacy.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


import qbitcoin.generated.qbit_pb2 as qbit__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10qbitlegacy.proto\x12\x03qrl\x1a\nqbit.proto\"\xf2\x08\n\rLegacyMessage\x12.\n\tfunc_name\x18\x01 \x01(\x0e\x32\x1b.qrl.LegacyMessage.FuncName\x12\x1d\n\x06noData\x18\x02 \x01(\x0b\x32\x0b.qrl.NoDataH\x00\x12\x1d\n\x06veData\x18\x03 \x01(\x0b\x32\x0b.qrl.VEDataH\x00\x12\x1d\n\x06plData\x18\x04 \x01(\x0b\x32\x0b.qrl.PLDataH\x00\x12!\n\x08pongData\x18\x05 \x01(\x0b\x32\r.qrl.PONGDataH\x00\x12\x1d\n\x06mrData\x18\x06 \x01(\x0b\x32\x0b.qrl.MRDataH\x00\x12\x1b\n\x05\x62lock\x18\x07 \x01(\x0b\x32\n.qrl.BlockH\x00\x12\x1d\n\x06\x66\x62\x44\x61ta\x18\x08 \x01(\x0b\x32\x0b.qrl.FBDataH\x00\x12\x1d\n\x06pbData\x18\t \x01(\x0b\x32\x0b.qrl.PBDataH\x00\x12&\n\x06\x62hData\x18\n \x01(\x0b\x32\x14.qrl.BlockHeightDataH\x00\x12\"\n\x06txData\x18\x0b \x01(\x0b\x32\x10.qrl.TransactionH\x00\x12\"\n\x06mtData\x18\x0c \x01(\x0b\x32\x10.qrl.TransactionH\x00\x12\"\n\x06tkData\x18\r \x01(\x0b\x32\x10.qrl.TransactionH\x00\x12\"\n\x06ttData\x18\x0e \x01(\x0b\x32\x10.qrl.TransactionH\x00\x12\"\n\x06ltData\x18\x0f \x01(\x0b\x32\x10.qrl.TransactionH\x00\x12\"\n\x06slData\x18\x10 \x01(\x0b\x32\x10.qrl.TransactionH\x00\x12\x31\n\x07\x65phData\x18\x11 \x01(\x0b\x32\x1e.qrl.EncryptedEphemeralMessageH\x00\x12!\n\x08syncData\x18\x12 \x01(\x0b\x32\r.qrl.SYNCDataH\x00\x12-\n\x0e\x63hainStateData\x18\x13 \x01(\x0b\x32\x13.qrl.NodeChainStateH\x00\x12-\n\x0enodeHeaderHash\x18\x14 \x01(\x0b\x32\x13.qrl.NodeHeaderHashH\x00\x12-\n\np2pAckData\x18\x15 \x01(\x0b\x32\x17.qrl.P2PAcknowledgementH\x00\x12\"\n\x06mcData\x18\x16 \x01(\x0b\x32\x10.qrl.TransactionH\x00\x12\"\n\x06msData\x18\x17 \x01(\x0b\x32\x10.qrl.TransactionH\x00\x12\"\n\x06mvData\x18\x18 \x01(\x0b\x32\x10.qrl.TransactionH\x00\"\xdf\x01\n\x08\x46uncName\x12\x06\n\x02VE\x10\x00\x12\x06\n\x02PL\x10\x01\x12\x08\n\x04PONG\x10\x02\x12\x06\n\x02MR\x10\x03\x12\x07\n\x03SFM\x10\x04\x12\x06\n\x02\x42K\x10\x05\x12\x06\n\x02\x46\x42\x10\x06\x12\x06\n\x02PB\x10\x07\x12\x06\n\x02\x42H\x10\x08\x12\x06\n\x02TX\x10\t\x12\x06\n\x02LT\x10\n\x12\x07\n\x03\x45PH\x10\x0b\x12\x06\n\x02MT\x10\x0c\x12\x06\n\x02TK\x10\r\x12\x06\n\x02TT\x10\x0e\x12\x06\n\x02SL\x10\x0f\x12\x08\n\x04SYNC\x10\x10\x12\x0e\n\nCHAINSTATE\x10\x11\x12\x10\n\x0cHEADERHASHES\x10\x12\x12\x0b\n\x07P2P_ACK\x10\x13\x12\x06\n\x02MC\x10\x14\x12\x06\n\x02MS\x10\x15\x12\x06\n\x02MV\x10\x16\x42\x06\n\x04\x64\x61ta\"\x08\n\x06NoData\"H\n\x06VEData\x12\x0f\n\x07version\x18\x01 \x01(\t\x12\x19\n\x11genesis_prev_hash\x18\x02 \x01(\x0c\x12\x12\n\nrate_limit\x18\x03 \x01(\x04\"/\n\x06PLData\x12\x10\n\x08peer_ips\x18\x01 \x03(\t\x12\x13\n\x0bpublic_port\x18\x02 \x01(\r\"\n\n\x08PONGData\"\x9d\x01\n\x06MRData\x12\x0c\n\x04hash\x18\x01 \x01(\x0c\x12)\n\x04type\x18\x02 \x01(\x0e\x32\x1b.qrl.LegacyMessage.FuncName\x12\x16\n\x0estake_selector\x18\x03 \x01(\x0c\x12\x14\n\x0c\x62lock_number\x18\x04 \x01(\x04\x12\x17\n\x0fprev_headerhash\x18\x05 \x01(\x0c\x12\x13\n\x0breveal_hash\x18\x06 \x01(\x0c\"@\n\x06\x42KData\x12\x1b\n\x06mrData\x18\x01 \x01(\x0b\x32\x0b.qrl.MRData\x12\x19\n\x05\x62lock\x18\x02 \x01(\x0b\x32\n.qrl.Block\"\x17\n\x06\x46\x42\x44\x61ta\x12\r\n\x05index\x18\x01 \x01(\x04\"#\n\x06PBData\x12\x19\n\x05\x62lock\x18\x01 \x01(\x0b\x32\n.qrl.Block\"\x19\n\x08SYNCData\x12\r\n\x05state\x18\x01 \x01(\tb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'qbitlegacy_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_LEGACYMESSAGE']._serialized_start=38
  _globals['_LEGACYMESSAGE']._serialized_end=1176
  _globals['_LEGACYMESSAGE_FUNCNAME']._serialized_start=945
  _globals['_LEGACYMESSAGE_FUNCNAME']._serialized_end=1168
  _globals['_NODATA']._serialized_start=1178
  _globals['_NODATA']._serialized_end=1186
  _globals['_VEDATA']._serialized_start=1188
  _globals['_VEDATA']._serialized_end=1260
  _globals['_PLDATA']._serialized_start=1262
  _globals['_PLDATA']._serialized_end=1309
  _globals['_PONGDATA']._serialized_start=1311
  _globals['_PONGDATA']._serialized_end=1321
  _globals['_MRDATA']._serialized_start=1324
  _globals['_MRDATA']._serialized_end=1481
  _globals['_BKDATA']._serialized_start=1483
  _globals['_BKDATA']._serialized_end=1547
  _globals['_FBDATA']._serialized_start=1549
  _globals['_FBDATA']._serialized_end=1572
  _globals['_PBDATA']._serialized_start=1574
  _globals['_PBDATA']._serialized_end=1609
  _globals['_SYNCDATA']._serialized_start=1611
  _globals['_SYNCDATA']._serialized_end=1636
# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python3
"""
Simple script to get Qbitcoin network difficulty using generated protobuf files
"""

import grpc
import sys
import os

# Add the qbitcoin package to path
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))

from qbitcoin.generated import qbit_pb2, qbit_pb2_grpc

def get_difficulty():
    """Get current network difficulty"""
    
    # Connect to Qbitcoin node
    channel = grpc.insecure_channel('134.122.79.166:19009')
    stub = qbit_pb2_grpc.PublicAPIStub(channel)
    
    try:
        # Create request with timeseries enabled
        request = qbit_pb2.GetStatsReq(include_timeseries=True)
        
        # Get stats
        response = stub.GetStats(request)
        
        # Extract latest difficulty
        if response.block_timeseries:
            latest_block = response.block_timeseries[-1]
            difficulty = latest_block.difficulty
            
            print(f"Current Difficulty: {difficulty}")
            print(f"Block Number: {latest_block.number}")
            print(f"Timestamp: {latest_block.timestamp}")
            
            return difficulty
        else:
            print("No timeseries data available")
            return None
            
    except grpc.RpcError as e:
        print(f"gRPC Error: {e}")
        return None
    except Exception as e:
        print(f"Error: {e}")
        return None
    finally:
        channel.close()

if __name__ == "__main__":
    difficulty = get_difficulty()
    if difficulty:
        print(f"\nDifficulty: {difficulty}")
    else:
        print("Failed to get difficulty")



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.


class Indexer:
    def __init__(self, name: bytes, db):
        self._name = name
        self._db = db
        self._data = dict()

    @property
    def data(self):
        return self._data

    def put(self, batch=None):
        for key, value in self._data.items():
            if value.delete:
                self._db.delete(self.generate_key(key), batch)
                continue

            self._db.put_raw(self.generate_key(key), value.SerializeToString(), batch)

    def load(self, key, meta_data) -> bool:
        generated_key = self.generate_key(key)
        try:
            meta_data.ParseFromString(self._db.get_raw(generated_key))
            self._data[key] = meta_data
        except Exception:
            return False

        return True

    def remove(self, batch=None):
        for key in self._data.items():
            self._db.delete(self.generate_key(key), batch)

    def generate_key(self, keys) -> bytes:
        if not isinstance(keys, tuple):
            raise Exception("Keys are not of type tuple for IndexerData")
        new_key = self._name

        for key in keys:
            if not isinstance(key, bytes):
                if not isinstance(key, str):
                    raise Exception("Invalid key datatype, neither bytes nor string")
                key = key.encode()
            new_key += b'_' + key

        return new_key



# -*- coding: utf-8 -*-
"""
Centralized version management for Qbitcoin
This is the single source of truth for version information
"""

__version__ = "1.1.3"

def get_version():
    """Get the current version"""
    return __version__

def get_version_info():
    """Get version as tuple"""
    return tuple(map(int, __version__.split('.')))

# For backwards compatibility with versioneer
def get_versions():
    """Compatibility function for versioneer"""
    return {
        'version': __version__,
        'full-revisionid': None,
        'dirty': False,
        'error': None,
        'date': None
    }



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.

from pyqrllib.pyqrllib import hstr2bin, bin2hstr

from qbitcoin.core.misc import logger
from qbitcoin.services.BaseService import BaseService
from qbitcoin.services.grpcHelper import GrpcExceptionWrapper
from qbitcoin.generated import qbit_pb2


class StakingAPIService(BaseService):
    def __init__(self, qbitnode):
        super().__init__(qbitnode)

    @GrpcExceptionWrapper(qbit_pb2.GetStakingInfoResp)
    def GetStakingInfo(self, request: qbit_pb2.GetStakingInfoReq, context) -> qbit_pb2.GetStakingInfoResp:
        """Get current staking information"""
        logger.debug("[StakingAPI] GetStakingInfo")
        answer = qbit_pb2.GetStakingInfoResp()
        
        try:
            staking_info = self.qbitnode.get_staking_info()
            
            answer.total_stakers = staking_info.get('total_stakers', 0)
            answer.active_stakers = staking_info.get('active_stakers', 0)
            answer.total_stake = staking_info.get('total_stake', 0)
            answer.donor_wallet = staking_info.get('donor_wallet', '')
            answer.last_reward_block = staking_info.get('last_reward_block', 0)
            answer.staking_enabled = staking_info.get('staking_enabled', False)
            
        except Exception as e:
            logger.error("GetStakingInfo failed: %s", str(e))
            answer.total_stakers = 0
            answer.active_stakers = 0
            answer.total_stake = 0
            answer.donor_wallet = ''
            answer.last_reward_block = 0
            answer.staking_enabled = False
            
        return answer

    @GrpcExceptionWrapper(qbit_pb2.GetStakersResp)
    def GetStakers(self, request: qbit_pb2.GetStakersReq, context) -> qbit_pb2.GetStakersResp:
        """Get list of active stakers"""
        logger.debug("[StakingAPI] GetStakers")
        answer = qbit_pb2.GetStakersResp()
        
        try:
            stakers = self.qbitnode.get_staker_list()
            
            for address_str, stake_info in stakers.items():
                staker_info = qbit_pb2.StakerInfo()
                staker_info.address = address_str
                staker_info.balance = stake_info.get('balance', 0)
                staker_info.stake_start_time = int(stake_info.get('stake_start_time', 0))
                staker_info.last_seen_time = int(stake_info.get('last_seen_time', 0))
                staker_info.is_online = stake_info.get('is_online', False)
                
                answer.stakers.append(staker_info)
                
        except Exception as e:
            logger.error("GetStakers failed: %s", str(e))
            
        return answer

    @GrpcExceptionWrapper(qbit_pb2.StartStakingResp)
    def StartStaking(self, request: qbit_pb2.StartStakingReq, context) -> qbit_pb2.StartStakingResp:
        """Start staking for an address"""
        logger.debug("[StakingAPI] StartStaking")
        answer = qbit_pb2.StartStakingResp()
        
        try:
            # Validate address format
            if not request.address.startswith('Q'):
                answer.success = False
                answer.error_message = "Address must start with 'Q'"
                return answer
                
            # Convert address
            address_bytes = bytes(hstr2bin(request.address[1:]))
            
            # Start staking
            success = self.qbitnode.start_staking(address_bytes)
            
            answer.success = success
            if not success:
                answer.error_message = "Failed to start staking. Check balance and requirements."
            else:
                answer.error_message = ""
                
        except Exception as e:
            logger.error("StartStaking failed: %s", str(e))
            answer.success = False
            answer.error_message = str(e)
            
        return answer

    @GrpcExceptionWrapper(qbit_pb2.StopStakingResp)
    def StopStaking(self, request: qbit_pb2.StopStakingReq, context) -> qbit_pb2.StopStakingResp:
        """Stop staking for an address"""
        logger.debug("[StakingAPI] StopStaking")
        answer = qbit_pb2.StopStakingResp()
        
        try:
            # Validate address format
            if not request.address.startswith('Q'):
                answer.success = False
                answer.error_message = "Address must start with 'Q'"
                return answer
                
            # Convert address
            address_bytes = bytes(hstr2bin(request.address[1:]))
            
            # Stop staking
            success = self.qbitnode.stop_staking(address_bytes)
            
            answer.success = success
            if not success:
                answer.error_message = "Failed to stop staking. Address may not be staking."
            else:
                answer.error_message = ""
                
        except Exception as e:
            logger.error("StopStaking failed: %s", str(e))
            answer.success = False
            answer.error_message = str(e)
            
        return answer

    @GrpcExceptionWrapper(qbit_pb2.SetDonorWalletResp)
    def SetDonorWallet(self, request: qbit_pb2.SetDonorWalletReq, context) -> qbit_pb2.SetDonorWalletResp:
        """Set donor wallet for staking rewards"""
        logger.debug("[StakingAPI] SetDonorWallet")
        answer = qbit_pb2.SetDonorWalletResp()
        
        try:
            # Validate address format
            if not request.address.startswith('Q'):
                answer.success = False
                answer.error_message = "Address must start with 'Q'"
                return answer
                
            # Convert address
            address_bytes = bytes(hstr2bin(request.address[1:]))
            
            # Set donor wallet
            success = self.qbitnode.set_donor_wallet(address_bytes)
            
            answer.success = success
            if not success:
                answer.error_message = "Failed to set donor wallet."
            else:
                answer.error_message = ""
                
        except Exception as e:
            logger.error("SetDonorWallet failed: %s", str(e))
            answer.success = False
            answer.error_message = str(e)
            
        return answer

    @GrpcExceptionWrapper(qbit_pb2.StakerHeartbeatResp)
    def StakerHeartbeat(self, request: qbit_pb2.StakerHeartbeatReq, context) -> qbit_pb2.StakerHeartbeatResp:
        """Update staker heartbeat to maintain online status"""
        logger.debug("[StakingAPI] StakerHeartbeat")
        answer = qbit_pb2.StakerHeartbeatResp()
        
        try:
            # Validate address format
            if not request.address.startswith('Q'):
                answer.success = False
                answer.error_message = "Address must start with 'Q'"
                return answer
                
            # Convert address
            address_bytes = bytes(hstr2bin(request.address[1:]))
            
            # Update heartbeat
            self.qbitnode.update_staker_heartbeat(address_bytes)
            
            answer.success = True
            answer.error_message = ""
                
        except Exception as e:
            logger.error("StakerHeartbeat failed: %s", str(e))
            answer.success = False
            answer.error_message = str(e)
            
        return answer



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.

from pyqrllib.pyqrllib import bin2hstr

from qbitcoin.core import config
from qbitcoin.core.State import State
from qbitcoin.core.OptimizedAddressState import OptimizedAddressState
from qbitcoin.generated import qbit_pb2


class AddressState(object):
    def __init__(self, protobuf_block=None):
        self._data = protobuf_block
        if protobuf_block is None:
            self._data = qbit_pb2.AddressState()

    @property
    def pbdata(self):
        """
        Returns a protobuf object that contains persistable data representing this object
        :return: A protobuf AddressState object
        :rtype: qbit_pb2.AddressState
        """
        return self._data

    @property
    def address(self):
        return self._data.address

    @property
    def nonce(self):
        return self._data.nonce

    @property
    def balance(self):
        return self._data.balance

    @balance.setter
    def balance(self, new_balance: int):
        self._data.balance = new_balance

    @property
    def transaction_hashes(self):
        return self._data.transaction_hashes

    @property
    def latticePK_list(self):
        return self._data.latticePK_list

    @property
    def slave_pks_access_type(self):
        return self._data.slave_pks_access_type

    @staticmethod
    def create(address: bytes,
               nonce: int,
               balance: int,
               falcon_pk_list: list,
               tokens: dict,
               slave_pks_access_type: dict):
        address_state = AddressState()

        address_state._data.address = address
        address_state._data.nonce = nonce
        address_state._data.balance = balance
        
        if falcon_pk_list:
            address_state._data.falcon_pk_list.extend(falcon_pk_list)

        for token_txhash in tokens:
            address_state.update_token_balance(token_txhash, tokens[token_txhash])

        for slave_pk in slave_pks_access_type:
            address_state.add_slave_pks_access_type(slave_pk, slave_pks_access_type[slave_pk])

        return address_state

    def validate_slave_with_access_type(self, slave_pk: str, access_types: list):
        if slave_pk not in self.slave_pks_access_type:
            return False

        access_type = self.slave_pks_access_type[slave_pk]
        if access_type not in access_types:
            return False

        return True

    def update_token_balance(self, token_tx_hash: bytes, balance: int):
        str_token_tx_hash = bin2hstr(token_tx_hash)
        self._data.tokens[str_token_tx_hash] += balance
        if self._data.tokens[str_token_tx_hash] == 0:
            del self._data.tokens[str_token_tx_hash]

    def get_token_balance(self, token_tx_hash: bytes) -> int:
        str_token_tx_hash = bin2hstr(token_tx_hash)
        if str_token_tx_hash in self._data.tokens:
            return self._data.tokens[str_token_tx_hash]
        return 0

    def is_token_exists(self, token_tx_hash: bytes) -> bool:
        str_token_tx_hash = bin2hstr(token_tx_hash)
        if str_token_tx_hash in self._data.tokens:
            return True
        return False

    def add_slave_pks_access_type(self, slave_pk: bytes, access_type: int):
        self._data.slave_pks_access_type[str(slave_pk)] = access_type

    def remove_slave_pks_access_type(self, slave_pk: bytes):
        del self._data.slave_pks_access_type[str(slave_pk)]

    def add_lattice_pk(self, lattice_txn):
        lattice_pk = qbit_pb2.LatticePK(txhash=lattice_txn.txhash,
                                       dilithium_pk=lattice_txn.dilithium_pk,
                                       kyber_pk=lattice_txn.kyber_pk)

        self._data.latticePK_list.extend([lattice_pk])

    def remove_lattice_pk(self, lattice_txn):
        for i, lattice_pk in enumerate(self._data.latticePK_list):
            if lattice_pk.txhash == lattice_txn.txhash:
                del self._data.latticePK_list[i]
                break

    def add_falcon_pk(self, public_key: bytes):
        """
        Add a Falcon 512 public key to the address state
        This is used to track used keys for preventing key reuse
        
        Args:
            public_key (bytes): A Falcon 512 public key to add
        """
        self._data.falcon_pk_list.append(public_key)

    def remove_falcon_pk(self, public_key: bytes):
        """
        Remove a Falcon 512 public key from the address state
        Used during rollbacks to undo transactions
        
        Args:
            public_key (bytes): Public key to remove
        """
        if public_key in self._data.falcon_pk_list:
            self._data.falcon_pk_list.remove(public_key)

    def is_falcon_pk_used(self, public_key: bytes) -> bool:
        """
        Check if a Falcon 512 public key has been used before
        
        Args:
            public_key (bytes): Public key to check
            
        Returns:
            bool: True if the key has been used previously
        """
        return public_key in self._data.falcon_pk_list

    def increase_nonce(self):
        self._data.nonce += 1

    def decrease_nonce(self):
        self._data.nonce -= 1

    def get_slave_permission(self, slave_pk) -> int:
        slave_pk_str = str(slave_pk)
        if slave_pk_str in self._data.slave_pks_access_type:
            return self._data.slave_pks_access_type[slave_pk_str]

        return -1

    @staticmethod
    def get_default(address):
        address_state = AddressState.create(address=address,                                            nonce=config.dev.default_nonce,
                                            balance=config.dev.default_account_balance,
                                            falcon_pk_list=[],
                                            tokens=dict(),
                                            slave_pks_access_type=dict())
        if address == config.dev.coinbase_address:
            address_state.balance = int(config.dev.max_coin_supply * config.dev.shor_per_quanta)
        return address_state
    
    @staticmethod
    def address_is_valid(address: bytes) -> bool:
        """
        Validate Falcon address format only
        """
        return OptimizedAddressState.address_is_valid(address)

    def serialize(self):
        return self._data.SerializeToString()

    @staticmethod
    def put_addresses_state(state: State, addresses_state: dict, batch=None):
        """
        :param addresses_state:
        :param batch:
        :return:
        """
        for address in addresses_state:
            address_state = addresses_state[address]
            AddressState.put_address_state(state, address_state, batch)

    @staticmethod
    def put_address_state(state: State, address_state, batch=None):
        data = address_state.pbdata.SerializeToString()
        state._db.put_raw(address_state.address, data, batch)

    @staticmethod
    def get_address_state(state: State, address: bytes):
        try:
            data = state._db.get_raw(address)
            pbdata = qbit_pb2.AddressState()
            pbdata.ParseFromString(bytes(data))
            address_state = AddressState(pbdata)
            return address_state
        except KeyError:
            return AddressState.get_default(address)

    @staticmethod
    def return_all_addresses(state: State) -> list:
        addresses = []
        for key, data in state._db.db:
            if key[0] != b'Q':
                continue
            pbdata = qbit_pb2.AddressState()
            pbdata.ParseFromString(bytes(data))
            address_state = AddressState(pbdata)
            addresses.append(address_state)
        return addresses



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from concurrent.futures import ThreadPoolExecutor

import argparse
import os
import logging
import grpc
from typing import Optional
from time import sleep
from daemonize import Daemonize

from pyqrllib.pyqrllib import hstr2bin, mnemonic2bin, bin2hstr

from qbitcoin.core import config
from qbitcoin.core.AddressState import AddressState
from qbitcoin.core.FalconHelper import falcon_pk_to_address
from qbitcoin.daemon.helper import logger
from qbitcoin.daemon.helper.DaemonHelper import WalletDecryptionError, Wallet, UNRESERVED_OTS_INDEX_START
from qbitcoin.services.WalletAPIService import WalletAPIService
from qbitcoin.generated import qbit_pb2, qbit_pb2_grpc, qbitwallet_pb2
from qbitcoin.generated.qbitwallet_pb2_grpc import add_WalletAPIServicer_to_server
from qbitcoin.core.txs.TransferTransaction import TransferTransaction
from qbitcoin.core.txs.MessageTransaction import MessageTransaction
from qbitcoin.core.txs.SlaveTransaction import SlaveTransaction
from qbitcoin.core.txs.TokenTransaction import TokenTransaction
from qbitcoin.core.txs.TransferTokenTransaction import TransferTokenTransaction

CONNECTION_TIMEOUT = 30
config.create_path(config.user.wallet_dir)
pid = os.path.join(config.user.wallet_dir, 'qrl_walletd.pid')


class WalletD:
    def __init__(self):
        self._wallet_path = os.path.join(config.user.wallet_dir, 'walletd.json')
        self._public_stub = qbit_pb2_grpc.PublicAPIStub(grpc.insecure_channel(config.user.public_api_server))
        self._wallet = None
        self._passphrase = None
        self.load_wallet()

    def to_plain_blocks(self, block):
        pheader = qbitwallet_pb2.PlainBlockHeader()
        pheader.hash_header = bin2hstr(block.header.hash_header)
        pheader.block_number = block.header.block_number
        pheader.timestamp_seconds = block.header.timestamp_seconds
        pheader.hash_header_prev = bin2hstr(block.header.hash_header_prev)
        pheader.reward_block = block.header.reward_block
        pheader.reward_fee = block.header.reward_fee
        pheader.merkle_root = bin2hstr(block.header.merkle_root)

        pheader.mining_nonce = block.header.mining_nonce
        pheader.extra_nonce = block.header.extra_nonce

        pblock = qbitwallet_pb2.PlainBlock()
        pblock.header.MergeFrom(pheader)

        for tx in block.transactions:
            pblock.transactions.extend([self.to_plain_transaction(tx)])

        for genesis_balance in block.genesis_balance:
            pgb = qbitwallet_pb2.PlainGenesisBalance()
            pgb.address = self.address_to_qaddress(genesis_balance.address)
            pgb.balance = genesis_balance.balance
            pblock.genesis_balance.extend([pgb])

        return pblock

    @staticmethod
    def to_plain_address_amount(address_amount):
        am = qbitwallet_pb2.PlainAddressAmount()
        am.address = bin2hstr(address_amount.address)
        am.amount = address_amount.amount
        return am

    def to_plain_transaction(self, tx):
        ptx = qbitwallet_pb2.PlainTransaction()
        if not tx.WhichOneof('transactionType'):
            return ptx
        if tx.master_addr:
            ptx.master_addr = self.address_to_qaddress(tx.master_addr)
        ptx.fee = tx.fee
        ptx.public_key = bin2hstr(tx.public_key)
        ptx.signature = bin2hstr(tx.signature)
        ptx.nonce = tx.nonce
        ptx.transaction_hash = bin2hstr(tx.transaction_hash)
        if tx.WhichOneof('transactionType') != 'coinbase':
            ptx.signer_addr = self.get_address_from_pk(ptx.public_key)

        if tx.WhichOneof('transactionType') == "transfer":
            ptx.transfer.amounts.extend(tx.transfer.amounts)
            for addr in tx.transfer.addrs_to:
                ptx.transfer.addrs_to.extend([self.address_to_qaddress(addr)])

        elif tx.WhichOneof('transactionType') == 'coinbase':
            ptx.coinbase.addr_to = self.address_to_qaddress(tx.coinbase.addr_to)
            ptx.coinbase.amount = tx.coinbase.amount

        elif tx.WhichOneof('transactionType') == 'lattice_public_key':
            ptx.lattice_public_key.MergeFrom(ptx.lattice_public_key())
            ptx.lattice_public_key.kyber_pk = bin2hstr(tx.lattice_public_key.kyber_pk)
            ptx.lattice_public_key.dilithium_pk = bin2hstr(tx.lattice_public_key.dilithium_pk)

        elif tx.WhichOneof('transactionType') == 'message':
            ptx.message.message_hash = bin2hstr(tx.message.message_hash)

        elif tx.WhichOneof('transactionType') == 'token':
            try:
                ptx.token.symbol = tx.token.symbol.decode()
            except UnicodeDecodeError:
                ptx.token.symbol = bin2hstr(tx.token.symbol)
            except Exception:
                ptx.token.symbol = str(tx.token.symbol)
            try:
                ptx.token.name = tx.token.name.decode()
            except UnicodeDecodeError:
                ptx.token.name = bin2hstr(tx.token.name)
            except Exception:
                ptx.token.name = str(tx.token.name)

            ptx.token.owner = self.address_to_qaddress(tx.token.owner)
            ptx.token.decimals = tx.token.decimals
            for initial_balance in tx.token.initial_balances:
                ptx.token.initial_balances.extend([self.to_plain_address_amount(initial_balance)])

        elif tx.WhichOneof('transactionType') == 'transfer_token':
            ptx.transfer_token.token_txhash = bin2hstr(tx.transfer_token.token_txhash)
            ptx.transfer_token.addrs_to.extend(self.addresses_to_qaddress(tx.transfer_token.addrs_to))
            ptx.transfer_token.amounts.extend(tx.transfer_token.amounts)

        elif tx.WhichOneof('transactionType') == 'slave':
            for slave_pk in tx.slave.slave_pks:
                ptx.slave.slave_pks.extend([bin2hstr(slave_pk)])
            ptx.slave.access_types.extend(tx.slave.access_types)

        return ptx

    def generate_slave_tx(self, signer_pk: bytes, slave_pk_list: list, master_addr=None):
        return SlaveTransaction.create(slave_pks=slave_pk_list,
                                       access_types=[0] * len(slave_pk_list),
                                       fee=0,
                                       xmss_pk=signer_pk,
                                       master_addr=master_addr)

    def load_wallet(self):
        self._wallet = Wallet(self._wallet_path)

    @staticmethod
    def address_to_qaddress(address: bytes):
        return 'Q' + bin2hstr(address)

    @staticmethod
    def addresses_to_qaddress(addresses: list):
        qaddresses = []
        for address in addresses:
            qaddresses.append(WalletD.address_to_qaddress(address))
        return qaddresses

    @staticmethod
    def qaddress_to_address(qaddress: str) -> bytes:
        if not qaddress:
            return qaddress
        return bytes(hstr2bin(qaddress[1:]))

    @staticmethod
    def qaddresses_to_address(qaddresses: list) -> list:
        if not qaddresses:
            return qaddresses
        addresses = []
        for qaddress in qaddresses:
            addresses.append(WalletD.qaddress_to_address(qaddress))
        return addresses

    def authenticate(self):
        if not self._passphrase:
            if self._wallet.is_encrypted():
                raise ValueError('Failed: Passphrase Missing')

    def _encrypt_last_item(self):
        if not self._passphrase:
            return
        self._wallet.encrypt_item(len(self._wallet.address_items) - 1, self._passphrase)

    def _get_wallet_index_xmss(self, signer_address: str, ots_index: int):
        index, _ = self._wallet.get_address_item(signer_address)
        if index is None:
            raise Exception("Signer Address Not Found ", signer_address)
        xmss = self._wallet.get_xmss_by_index(index, self._passphrase)
        if ots_index > 0:
            xmss.set_ots_index(ots_index)
        return index, xmss

    def get_pk_list_from_xmss_list(self, slave_xmss_list):
        return [xmss.pk for xmss in slave_xmss_list]

    def add_new_address(self, height=10, hash_function='shake128') -> str:
        self.authenticate()

        if not hash_function:
            hash_function = 'shake128'

        if not height:
            height = 10

        self._wallet.add_new_address(height, hash_function, True)
        self._encrypt_last_item()
        self._wallet.save()
        logger.info("Added New Address")
        return self._wallet.address_items[-1].qaddress

    def add_new_address_with_slaves(self,
                                    height=10,
                                    number_of_slaves=config.user.number_of_slaves,
                                    hash_function='shake128') -> str:
        self.authenticate()

        if not hash_function:
            hash_function = 'shake128'

        if not height:
            height = 10

        if height < 6:
            raise Exception("Height cannot be less than 6")

        if not number_of_slaves:
            number_of_slaves = config.user.number_of_slaves

        if number_of_slaves > 100:
            raise Exception("Number of slaves cannot be more than 100")

        xmss = self._wallet.add_new_address(height, hash_function, True)
        slave_xmss_list = self._wallet.add_slave(index=-1,
                                                 height=height,
                                                 number_of_slaves=number_of_slaves,
                                                 hash_function=hash_function,
                                                 force=True)
        self._encrypt_last_item()

        slave_pk_list = self.get_pk_list_from_xmss_list(slave_xmss_list)
        slave_tx = self.generate_slave_tx(xmss.pk, slave_pk_list)
        self.sign_and_push_transaction(slave_tx, xmss, -1)

        self._wallet.save()
        logger.info("Added New Address With Slaves")
        return self._wallet.address_items[-1].qaddress

    def add_address_from_seed(self, seed=None) -> str:
        self.authenticate()

        words = seed.split()
        if len(words) == 34:
            bin_seed = mnemonic2bin(seed)
        elif len(seed) == 102:
            bin_seed = hstr2bin(seed)
        else:
            raise ValueError("Invalid Seed")

        address_from_seed = XMSS.from_extended_seed(bin_seed)
        if self._wallet.get_xmss_by_qaddress(address_from_seed.qaddress, self._passphrase):
            raise Exception("Address is already in the wallet")
        self._wallet.append_xmss(address_from_seed)
        self._encrypt_last_item()
        self._wallet.save()

        return address_from_seed.qaddress

    def list_address(self) -> list:
        self.authenticate()

        addresses = []
        for item in self._wallet.address_items:
            addresses.append(item.qaddress)

        return addresses

    def remove_address(self, qaddress: str) -> bool:
        self.authenticate()
        if self._wallet.remove(qaddress):
            logger.info("Removed Address %s", qaddress)
            return True

        return False

    def validate_address(self, qaddress: str) -> bool:
        try:
            return AddressState.address_is_valid(bytes(hstr2bin(qaddress[1:])))
        except Exception:
            return False

    def get_recovery_seeds(self, qaddress: str):
        self.authenticate()

        xmss = self._wallet.get_xmss_by_qaddress(qaddress, self._passphrase)
        if xmss:
            logger.info("Recovery seeds requested for %s", qaddress)
            return xmss.hexseed, xmss.mnemonic

        raise ValueError("No such address found in wallet")

    def get_wallet_info(self):
        self.authenticate()

        return self._wallet.wallet_info()

    def get_address_state(self, qaddress: str) -> AddressState:
        request = qbit_pb2.GetAddressStateReq(address=bytes(hstr2bin(qaddress[1:])))

        resp = self._public_stub.GetAddressState(request=request)
        return AddressState(resp.state)

    def sign_and_push_transaction(self,
                                  tx,
                                  xmss,
                                  index,
                                  group_index=None,
                                  slave_index=None,
                                  enable_save=True):
        logger.info("Signing %s transaction by %s | OTS index %s", tx.type, xmss.qaddress, xmss.ots_index)
        tx.sign(xmss)
        if not tx.validate(True):
            raise Exception("Invalid Transaction")

        if enable_save:
            if slave_index is None:  # noqa
                self._wallet.set_ots_index(index, xmss.ots_index)  # Move to next OTS index before broadcasting txn
            else:
                self._wallet.set_slave_ots_index(index, group_index, slave_index, xmss.ots_index)

        push_transaction_req = qbit_pb2.PushTransactionReq(transaction_signed=tx.pbdata)
        push_transaction_resp = self._public_stub.PushTransaction(push_transaction_req, timeout=CONNECTION_TIMEOUT)
        if push_transaction_resp.error_code != qbit_pb2.PushTransactionResp.SUBMITTED:
            raise Exception(push_transaction_resp.error_description)

    def try_txn_with_last_slave(self, item, index, group_index, xmss=None):
        slave = item.slaves[group_index][-1]

        # Ignore usage of last 5 ots indexes for the last slave in slave group
        if slave.index >= 2 ** slave.height - 5:
            return None

        slave_index = len(item.slaves[group_index]) - 1
        slave_address_state = self.get_address_state(slave.qaddress)

        ots_index = slave_address_state.get_unused_ots_index(slave.index)

        if ots_index is None:  # noqa
            self._wallet.set_slave_ots_index(index,
                                             group_index,
                                             slave_index,
                                             2 ** slave.height)
            return None
        if not xmss:
            target_address_item = slave
            if self._passphrase:
                target_address_item = self._wallet.decrypt_address_item(slave, self._passphrase)
            xmss = self._wallet.get_xmss_by_item(target_address_item, ots_index)
        else:
            xmss.set_ots_index(ots_index)

        return xmss

    def is_slave(self, master_address: bytes, slave_pk: bytes) -> bool:
        request = qbit_pb2.IsSlaveReq(master_address=master_address,
                                     slave_pk=slave_pk)

        resp = self._public_stub.IsSlave(request=request)
        return resp.result

    def get_unused_ots_index(self, address: bytes, unused_ots_index_from: int) -> Optional[int]:
        request = qbit_pb2.GetOTSReq(address=address,
                                    unused_ots_index_from=unused_ots_index_from)
        resp = self._public_stub.GetOTS(request=request)
        if not resp.unused_ots_index_found:
            return None
        return resp.next_unused_ots_index

    def get_slave(self, master_qaddress):
        index, item = self._wallet.get_address_item(master_qaddress)
        if index is None:
            raise Exception("Signer Address Not Found ", master_qaddress)

        # Should we check available OTS for master
        # Get slave list using address state
        address_state = self.get_address_state(master_qaddress)

        slave = item.slaves[-1][0]
        if not address_state.validate_slave_with_access_type(str(bytes(hstr2bin(slave.pk))), [0]):
            if len(item.slaves) == 1:
                qaddress = item.qaddress
                target_address_item = item
                group_index = None
            else:
                qaddress = item.slaves[-2][-1].qaddress
                target_address_item = item.slaves[-2][-1]
                group_index = -2

            address_state = self.get_address_state(qaddress)
            ots_index = address_state.get_unused_ots_index()

            if ots_index >= UNRESERVED_OTS_INDEX_START:
                raise Exception('Fatal Error!!! No reserved OTS index found')

            if self._passphrase:
                target_address_item = self._wallet.decrypt_address_item(target_address_item, self._passphrase)

            xmss = self._wallet.get_xmss_by_item(target_address_item, ots_index)

            slaves_pk = [bytes(hstr2bin(slave_item.pk)) for slave_item in item.slaves[-1]]
            tx = self.generate_slave_tx(xmss.pk,
                                        slaves_pk,
                                        self.qaddress_to_address(master_qaddress))

            self.sign_and_push_transaction(tx,
                                           xmss,
                                           index,
                                           enable_save=False)

            if len(item.slaves) > 1:
                if self.try_txn_with_last_slave(item, index, group_index, xmss):
                    return index, len(item.slaves) - 2, len(item.slaves[group_index]) - 1, xmss

        else:
            if len(item.slaves) > 1:
                group_index = len(item.slaves) - 2
                xmss = self.try_txn_with_last_slave(item, index, group_index)
                if xmss:
                    return index, group_index, len(item.slaves[group_index]) - 1, xmss
            group_index = len(item.slaves) - 1
            last_slaves = item.slaves[-1]
            for slave_index, slave in enumerate(last_slaves):
                # Check if all ots index has been marked as used
                if slave.index > 2 ** slave.height - 1:
                    continue

                # Ignore usage of last 5 ots indexes for the last slave in slave group
                if slave_index + 1 == len(last_slaves) and slave.index >= 2 ** slave.height - 5:
                    continue

                if self._passphrase:
                    slave = self._wallet.decrypt_address_item(slave, self._passphrase)

                slave_address_state = self.get_address_state(slave.qaddress)

                if slave_index + 1 == len(last_slaves) and slave.index > 2 ** slave.height - 100:

                    ots_index = slave_address_state.get_unused_ots_index(0)
                    if ots_index >= UNRESERVED_OTS_INDEX_START:
                        raise Exception("Fatal Error, no unused reserved OTS index")

                    curr_slave_xmss = self._wallet.get_xmss_by_item(slave, ots_index)

                    slave_xmss_list = self._wallet.add_slave(index=index,
                                                             height=slave.height,
                                                             number_of_slaves=config.user.number_of_slaves,
                                                             passphrase=self._passphrase,
                                                             force=True)
                    slave_pk_list = self.get_pk_list_from_xmss_list(slave_xmss_list)

                    tx = self.generate_slave_tx(bytes(hstr2bin(slave.pk)),
                                                slave_pk_list,
                                                self.qaddress_to_address(item.qaddress))

                    self.sign_and_push_transaction(tx,
                                                   curr_slave_xmss,
                                                   index,
                                                   enable_save=False)

                ots_index = slave_address_state.get_unused_ots_index(slave.index)

                if ots_index is None:  # noqa
                    self._wallet.set_slave_ots_index(index,
                                                     group_index,
                                                     slave_index,
                                                     2 ** slave.height)
                    continue

                slave_xmss = self._wallet.get_xmss_by_item(slave, ots_index)

                return index, group_index, slave_index, slave_xmss

        return index, -1, -1, None

    # def get_slave(self, master_qaddress):
    #     index, item = self._wallet.get_address_item(master_qaddress)
    #     if index is None:
    #         raise Exception("Signer Address Not Found ", master_qaddress)
    #
    #     # Should we check available OTS for master
    #     # Get slave list using address state
    #     address_state = self.get_address_state(master_qaddress)
    #
    #     slave = item.slaves[-1][0]
    #     if not address_state.validate_slave_with_access_type(str(bytes(hstr2bin(slave.pk))), [0]):
    #         if len(item.slaves) == 1:
    #             qaddress = item.qaddress
    #             target_address_item = item
    #             group_index = None
    #         else:
    #             qaddress = item.slaves[-2][-1].qaddress
    #             target_address_item = item.slaves[-2][-1]
    #             group_index = -2
    #
    #         address_state = self.get_address_state(qaddress)
    #         ots_index = address_state.get_unused_ots_index()
    #
    #         if ots_index >= UNRESERVED_OTS_INDEX_START:
    #             raise Exception('Fatal Error!!! No reserved OTS index found')
    #
    #         if self._passphrase:
    #             target_address_item = self._wallet.decrypt_address_item(target_address_item, self._passphrase)
    #
    #         xmss = self._wallet.get_xmss_by_item(target_address_item, ots_index)
    #
    #         slaves_pk = [bytes(hstr2bin(slave_item.pk)) for slave_item in item.slaves[-1]]
    #         tx = self.generate_slave_tx(xmss.pk,
    #                                     slaves_pk,
    #                                     self.qaddress_to_address(master_qaddress))
    #
    #         self.sign_and_push_transaction(tx,
    #                                        xmss,
    #                                        index,
    #                                        enable_save=False)
    #
    #         if len(item.slaves) > 1:
    #             if self.try_txn_with_last_slave(item, index, group_index, xmss):
    #                 return index, len(item.slaves) - 2, len(item.slaves[group_index]) - 1, xmss
    #
    #     else:
    #         if len(item.slaves) > 1:
    #             group_index = len(item.slaves) - 2
    #             xmss = self.try_txn_with_last_slave(item, index, group_index)
    #             if xmss:
    #                 return index, group_index, len(item.slaves[group_index]) - 1, xmss
    #         group_index = len(item.slaves) - 1
    #         last_slaves = item.slaves[-1]
    #         for slave_index in range(len(last_slaves)):
    #             slave = last_slaves[slave_index]
    #
    #             # Check if all ots index has been marked as used
    #             if slave.index > 2 ** slave.height - 1:
    #                 continue
    #
    #             # Ignore usage of last 5 ots indexes for the last slave in slave group
    #             if slave_index + 1 == len(last_slaves) and slave.index >= 2 ** slave.height - 5:
    #                 continue
    #
    #             if self._passphrase:
    #                 slave = self._wallet.decrypt_address_item(slave, self._passphrase)
    #
    #             slave_address_state = self.get_address_state(slave.qaddress)
    #
    #             if slave_index + 1 == len(last_slaves) and slave.index > 2 ** slave.height - 100:
    #
    #                 ots_index = slave_address_state.get_unused_ots_index(0)
    #                 if ots_index >= UNRESERVED_OTS_INDEX_START:
    #                     raise Exception("Fatal Error, no unused reserved OTS index")
    #
    #                 curr_slave_xmss = self._wallet.get_xmss_by_item(slave, ots_index)
    #
    #                 slave_xmss_list = self._wallet.add_slave(index=index,
    #                                                          height=slave.height,
    #                                                          number_of_slaves=config.user.number_of_slaves,
    #                                                          passphrase=self._passphrase,
    #                                                          force=True)
    #                 slave_pk_list = self.get_pk_list_from_xmss_list(slave_xmss_list)
    #
    #                 tx = self.generate_slave_tx(bytes(hstr2bin(slave.pk)),
    #                                             slave_pk_list,
    #                                             self.qaddress_to_address(item.qaddress))
    #
    #                 self.sign_and_push_transaction(tx,
    #                                                curr_slave_xmss,
    #                                                index,
    #                                                enable_save=False)
    #
    #             ots_index = slave_address_state.get_unused_ots_index(slave.index)
    #
    #             if ots_index == None:  # noqa
    #                 self._wallet.set_slave_ots_index(index,
    #                                                  group_index,
    #                                                  slave_index,
    #                                                  2 ** slave.height)
    #                 continue
    #
    #             slave_xmss = self._wallet.get_xmss_by_item(slave, ots_index)
    #
    #             return index, group_index, slave_index, slave_xmss
    #
    #     return index, -1, -1, None

    def get_slave_xmss(self, master_qaddress):
        index, group_index, slave_index, slave_xmss = self.get_slave(master_qaddress)

        return index, group_index, slave_index, slave_xmss

    def get_slave_list(self, qaddress) -> list:
        self.authenticate()
        _, addr_item = self._wallet.get_address_item(qaddress)
        if addr_item is None:
            raise Exception("Address Not Found ", qaddress)
        return addr_item.slaves

    def verify_ots(self, signer_address, xmss, user_ots_index):
        addr_state = self.get_address_state(signer_address)

        verified_ots_index = self.get_unused_ots_index(addr_state.address, xmss.ots_index)

        if verified_ots_index is None:  # noqa
            raise Exception("No Unused OTS key found")

        if user_ots_index > 0:
            if verified_ots_index != xmss.ots_index:
                raise Exception("Used OTS Index %s", user_ots_index)
        else:
            xmss.set_ots_index(verified_ots_index)

    def relay_transfer_txn(self,
                           qaddresses_to: list,
                           amounts: list,
                           fee: int,
                           master_qaddress,
                           signer_address: str,
                           ots_index: int):
        self.authenticate()
        index, xmss = self._get_wallet_index_xmss(signer_address, ots_index)
        self.verify_ots(signer_address, xmss, user_ots_index=ots_index)

        tx = TransferTransaction.create(addrs_to=self.qaddresses_to_address(qaddresses_to),
                                        amounts=amounts,
                                        message_data=None,
                                        fee=fee,
                                        xmss_pk=xmss.pk,
                                        master_addr=self.qaddress_to_address(master_qaddress))

        self.sign_and_push_transaction(tx, xmss, index)

        return self.to_plain_transaction(tx.pbdata)

    def relay_transfer_txn_by_slave(self,
                                    qaddresses_to: list,
                                    amounts: list,
                                    fee: int,
                                    master_qaddress):
        self.authenticate()
        index, group_index, slave_index, slave_xmss = self.get_slave_xmss(master_qaddress)
        if slave_index == -1:
            raise Exception("No Slave Found")

        tx = TransferTransaction.create(addrs_to=self.qaddresses_to_address(qaddresses_to),
                                        amounts=amounts,
                                        message_data=None,
                                        fee=fee,
                                        xmss_pk=slave_xmss.pk,
                                        master_addr=self.qaddress_to_address(master_qaddress))

        self.sign_and_push_transaction(tx, slave_xmss, index, group_index, slave_index)

        return self.to_plain_transaction(tx.pbdata)

    def relay_message_txn(self,
                          message: str,
                          fee: int,
                          master_qaddress,
                          signer_address: str,
                          ots_index: int):
        self.authenticate()
        index, xmss = self._get_wallet_index_xmss(signer_address, ots_index)
        self.verify_ots(signer_address, xmss, user_ots_index=ots_index)

        tx = MessageTransaction.create(message_hash=message.encode(),
                                       addr_to=None,
                                       fee=fee,
                                       xmss_pk=xmss.pk,
                                       master_addr=self.qaddress_to_address(master_qaddress))

        self.sign_and_push_transaction(tx, xmss, index)

        return self.to_plain_transaction(tx.pbdata)

    def relay_message_txn_by_slave(self,
                                   message: str,
                                   fee: int,
                                   master_qaddress):
        self.authenticate()
        index, group_index, slave_index, slave_xmss = self.get_slave_xmss(master_qaddress)
        if slave_index == -1:
            raise Exception("No Slave Found")

        tx = MessageTransaction.create(message_hash=message.encode(),
                                       addr_to=None,
                                       fee=fee,
                                       xmss_pk=slave_xmss.pk,
                                       master_addr=self.qaddress_to_address(master_qaddress))

        self.sign_and_push_transaction(tx, slave_xmss, index, group_index, slave_index)

        return self.to_plain_transaction(tx.pbdata)

    def relay_token_txn(self,
                        symbol: str,
                        name: str,
                        owner_qaddress: str,
                        decimals: int,
                        qaddresses: list,
                        amounts: list,
                        fee: int,
                        master_qaddress,
                        signer_address: str,
                        ots_index: int):
        self.authenticate()

        if len(qaddresses) != len(amounts):
            raise Exception("Number of Addresses & Amounts Mismatch")

        index, xmss = self._get_wallet_index_xmss(signer_address, ots_index)
        self.verify_ots(signer_address, xmss, user_ots_index=ots_index)

        initial_balances = []
        for idx, qaddress in enumerate(qaddresses):
            initial_balances.append(qbit_pb2.AddressAmount(address=self.qaddress_to_address(qaddress),
                                                          amount=amounts[idx]))
        tx = TokenTransaction.create(symbol=symbol.encode(),
                                     name=name.encode(),
                                     owner=self.qaddress_to_address(owner_qaddress),
                                     decimals=decimals,
                                     initial_balances=initial_balances,
                                     fee=fee,
                                     xmss_pk=xmss.pk,
                                     master_addr=self.qaddress_to_address(master_qaddress))

        self.sign_and_push_transaction(tx, xmss, index)

        return self.to_plain_transaction(tx.pbdata)

    def relay_token_txn_by_slave(self,
                                 symbol: str,
                                 name: str,
                                 owner_qaddress: str,
                                 decimals: int,
                                 qaddresses: list,
                                 amounts: list,
                                 fee: int,
                                 master_qaddress):
        self.authenticate()

        if len(qaddresses) != len(amounts):
            raise Exception("Number of Addresses & Amounts Mismatch")

        index, group_index, slave_index, slave_xmss = self.get_slave_xmss(master_qaddress)
        if slave_index == -1:
            raise Exception("No Slave Found")

        initial_balances = []
        for idx, qaddress in enumerate(qaddresses):
            initial_balances.append(qbit_pb2.AddressAmount(address=self.qaddress_to_address(qaddress),
                                                          amount=amounts[idx]))
        tx = TokenTransaction.create(symbol=symbol.encode(),
                                     name=name.encode(),
                                     owner=self.qaddress_to_address(owner_qaddress),
                                     decimals=decimals,
                                     initial_balances=initial_balances,
                                     fee=fee,
                                     xmss_pk=slave_xmss.pk,
                                     master_addr=self.qaddress_to_address(master_qaddress))

        self.sign_and_push_transaction(tx, slave_xmss, index, group_index, slave_index)

        return self.to_plain_transaction(tx.pbdata)

    def relay_transfer_token_txn(self,
                                 qaddresses_to: list,
                                 amounts: list,
                                 token_txhash: str,
                                 fee: int,
                                 master_qaddress,
                                 signer_address: str,
                                 ots_index: int):
        self.authenticate()
        index, xmss = self._get_wallet_index_xmss(signer_address, ots_index)
        self.verify_ots(signer_address, xmss, user_ots_index=ots_index)

        tx = TransferTokenTransaction.create(token_txhash=bytes(hstr2bin(token_txhash)),
                                             addrs_to=self.qaddresses_to_address(qaddresses_to),
                                             amounts=amounts,
                                             fee=fee,
                                             xmss_pk=xmss.pk,
                                             master_addr=self.qaddress_to_address(master_qaddress))

        self.sign_and_push_transaction(tx, xmss, index)

        return self.to_plain_transaction(tx.pbdata)

    def relay_transfer_token_txn_by_slave(self,
                                          qaddresses_to: list,
                                          amounts: list,
                                          token_txhash: str,
                                          fee: int,
                                          master_qaddress):
        self.authenticate()
        index, group_index, slave_index, slave_xmss = self.get_slave_xmss(master_qaddress)
        if slave_index == -1:
            raise Exception("No Slave Found")

        tx = TransferTokenTransaction.create(token_txhash=bytes(hstr2bin(token_txhash)),
                                             addrs_to=self.qaddresses_to_address(qaddresses_to),
                                             amounts=amounts,
                                             fee=fee,
                                             xmss_pk=slave_xmss.pk,
                                             master_addr=self.qaddress_to_address(master_qaddress))

        self.sign_and_push_transaction(tx, slave_xmss, index, group_index, slave_index)

        return self.to_plain_transaction(tx.pbdata)

    def relay_slave_txn(self,
                        slave_pks: list,
                        access_types: list,
                        fee: int,
                        master_qaddress,
                        signer_address: str,
                        ots_index: int):
        self.authenticate()
        index, xmss = self._get_wallet_index_xmss(signer_address, ots_index)
        self.verify_ots(signer_address, xmss, user_ots_index=ots_index)

        tx = SlaveTransaction.create(slave_pks=slave_pks,
                                     access_types=access_types,
                                     fee=fee,
                                     xmss_pk=xmss.pk,
                                     master_addr=self.qaddress_to_address(master_qaddress))

        self.sign_and_push_transaction(tx, xmss, index)

        return self.to_plain_transaction(tx.pbdata)

    def relay_slave_txn_by_slave(self,
                                 slave_pks: list,
                                 access_types: list,
                                 fee: int,
                                 master_qaddress):
        self.authenticate()
        index, group_index, slave_index, slave_xmss = self.get_slave_xmss(master_qaddress)
        if slave_index == -1:
            raise Exception("No Slave Found")

        tx = SlaveTransaction.create(slave_pks=slave_pks,
                                     access_types=access_types,
                                     fee=fee,
                                     xmss_pk=slave_xmss.pk,
                                     master_addr=self.qaddress_to_address(master_qaddress))

        self.sign_and_push_transaction(tx, slave_xmss, index, group_index, slave_index)

        return self.to_plain_transaction(tx.pbdata)

    def encrypt_wallet(self, passphrase: str):
        if self._wallet.is_encrypted():
            raise Exception('Wallet Already Encrypted')
        if not passphrase:
            raise Exception("Missing Passphrase")
        if len(self._wallet.address_items) == 0:
            raise ValueError('Cannot be encrypted as wallet does not have any address.')
        self._wallet.encrypt(passphrase)
        self._wallet.save()
        logger.info("Wallet Encrypted")

    def lock_wallet(self):
        if not self._wallet.is_encrypted():
            raise Exception('You cannot lock an unencrypted Wallet')

        self._passphrase = None
        logger.info("Wallet Locked")

    def unlock_wallet(self, passphrase: str):
        if not self._wallet.is_encrypted():
            raise Exception('You cannot unlock an unencrypted Wallet')

        self._passphrase = passphrase
        self._wallet.decrypt(passphrase, first_address_only=True)  # Check if Password Correct
        self._wallet.encrypt_item(0, passphrase)  # Re-Encrypt first address item
        logger.info("Wallet Unlocked")

    def change_passphrase(self, old_passphrase: str, new_passphrase: str):
        if len(old_passphrase) == 0:
            raise Exception('Missing Old Passphrase')

        if len(new_passphrase) == 0:
            raise Exception('Missing New Passphrase')

        if old_passphrase == new_passphrase:
            raise Exception('Old Passphrase and New Passphrase cannot be same')

        self._passphrase = old_passphrase

        if not self._wallet:
            self.unlock_wallet(old_passphrase)
        try:
            self._wallet.decrypt(old_passphrase)
        except WalletDecryptionError:
            raise ValueError('Invalid Old Passphrase')

        self._wallet.encrypt(new_passphrase)
        self._wallet.save()
        self.lock_wallet()
        logger.info("Passphrase Changed")

    def get_mini_transactions_by_address(self, qaddress: str, item_per_page: int, page_number: int) -> tuple:
        address = self.qaddress_to_address(qaddress)
        response = self._public_stub.GetMiniTransactionsByAddress(
            qbit_pb2.GetMiniTransactionsByAddressReq(address=address,
                                                    item_per_page=item_per_page,
                                                    page_number=page_number))
        return response.mini_transactions, response.balance

    def get_transaction(self, tx_hash: str):
        txhash = bytes(hstr2bin(tx_hash))
        response = self._public_stub.GetTransaction(qbit_pb2.GetTransactionReq(tx_hash=txhash))
        block_header_hash = None
        if response.block_header_hash:
            block_header_hash = bin2hstr(response.block_header_hash)
        return self.to_plain_transaction(response.tx), str(response.confirmations), response.block_number, block_header_hash

    def get_balance(self, qaddress: str) -> int:
        address = self.qaddress_to_address(qaddress)
        response = self._public_stub.GetBalance(qbit_pb2.GetBalanceReq(address=address))
        return response.balance

    def get_total_balance(self) -> int:
        self.authenticate()

        addresses = []
        for item in self._wallet.address_items:
            addresses.append(bytes(hstr2bin(item.qaddress[1:])))

        response = self._public_stub.GetTotalBalance(qbit_pb2.GetTotalBalanceReq(addresses=addresses))
        return response.balance
 
    def get_height(self) -> int:
        response = self._public_stub.GetHeight(qbit_pb2.GetHeightReq())
        return response.height

    def get_block(self, header_hash: str):
        headerhash = bytes(hstr2bin(header_hash))
        response = self._public_stub.GetBlock(qbit_pb2.GetBlockReq(header_hash=headerhash))
        return self.to_plain_blocks(response.block)

    def get_block_by_number(self, block_number: int):
        response = self._public_stub.GetBlockByNumber(qbit_pb2.GetBlockByNumberReq(block_number=block_number))
        return self.to_plain_blocks(response.block)

    def get_address_from_pk(self, pk: str) -> str:
        # Only handle Falcon-512 public keys (1793 bytes)
        pk_bytes = bytes(hstr2bin(pk))
        if len(pk_bytes) == 1793:
            address = falcon_pk_to_address(pk_bytes)
            return self.address_to_qaddress(address)
        else:
            # Invalid key size for Falcon-512
            return ""

    def get_node_info(self):
        return self._public_stub.GetNodeState(qbit_pb2.GetNodeStateReq())


def run():
    logger.initialize_default(force_console_output=True).setLevel(logging.INFO)
    file_handler = logger.log_to_file()
    file_handler.setLevel(logging.INFO)

    LOG_FORMAT_CUSTOM = '%(asctime)s| %(levelname)s : %(message)s'  # noqa

    logger.set_colors(False, LOG_FORMAT_CUSTOM)
    logger.set_unhandled_exception_handler()

    walletd = WalletD()  # noqa
    wallet_server = grpc.server(ThreadPoolExecutor(max_workers=config.user.wallet_api_threads),
                                maximum_concurrent_rpcs=config.user.wallet_api_max_concurrent_rpc)
    add_WalletAPIServicer_to_server(WalletAPIService(walletd), wallet_server)

    wallet_server.add_insecure_port("{0}:{1}".format(config.user.wallet_api_host,
                                                     config.user.wallet_api_port))
    wallet_server.start()

    logger.info("WalletAPIService Started")

    try:
        while True:
            sleep(60)
    except Exception:  # noqa
        wallet_server.stop(0)


def parse_arguments():
    parser = argparse.ArgumentParser(description='QRL node')
    parser.add_argument('--qrldir', '-d', dest='qrl_dir', default=config.user.qrl_dir,
                        help="Use a different directory for node data/configuration")
    parser.add_argument('--network-type', dest='network_type', choices=['mainnet', 'testnet'],
                        default='mainnet', required=False, help="Runs QRL Testnet Node")
    return parser.parse_args()


def main():
    args = parse_arguments()

    qrl_dir_post_fix = ''
    copy_files = []
    if args.network_type == 'testnet':
        qrl_dir_post_fix = '-testnet'
        package_directory = os.path.dirname(os.path.abspath(__file__))
        copy_files.append(os.path.join(package_directory, '../network/testnet/genesis.yml'))
        copy_files.append(os.path.join(package_directory, '../network/testnet/config.yml'))

    config.user.qrl_dir = os.path.expanduser(os.path.normpath(args.qrl_dir) + qrl_dir_post_fix)
    config.create_path(config.user.qrl_dir, copy_files)
    config.user.load_yaml(config.user.config_path)

    daemon = Daemonize(app="qrl_walletd", pid=pid, action=run)
    daemon.start()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: qbitmining.proto
# Protobuf Python Version: 6.30.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    30,
    0,
    '',
    'qbitmining.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


import qbitcoin.generated.qbit_pb2 as qbit__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10qbitmining.proto\x12\x03qrl\x1a\nqbit.proto\"-\n\x1bGetBlockMiningCompatibleReq\x12\x0e\n\x06height\x18\x01 \x01(\x04\"\'\n\x15GetLastBlockHeaderReq\x12\x0e\n\x06height\x18\x01 \x01(\x04\"p\n\x1cGetBlockMiningCompatibleResp\x12%\n\x0b\x62lockheader\x18\x01 \x01(\x0b\x32\x10.qrl.BlockHeader\x12)\n\rblockmetadata\x18\x02 \x01(\x0b\x32\x12.qrl.BlockMetaData\"|\n\x16GetLastBlockHeaderResp\x12\x12\n\ndifficulty\x18\x01 \x01(\x04\x12\x0e\n\x06height\x18\x02 \x01(\x04\x12\x11\n\ttimestamp\x18\x03 \x01(\x04\x12\x0e\n\x06reward\x18\x04 \x01(\x04\x12\x0c\n\x04hash\x18\x05 \x01(\t\x12\r\n\x05\x64\x65pth\x18\x06 \x01(\x04\"+\n\x11GetBlockToMineReq\x12\x16\n\x0ewallet_address\x18\x01 \x01(\x0c\"\x80\x01\n\x12GetBlockToMineResp\x12\x1a\n\x12\x62locktemplate_blob\x18\x01 \x01(\t\x12\x12\n\ndifficulty\x18\x02 \x01(\x04\x12\x0e\n\x06height\x18\x03 \x01(\x04\x12\x17\n\x0freserved_offset\x18\x04 \x01(\r\x12\x11\n\tseed_hash\x18\x05 \x01(\t\"#\n\x13SubmitMinedBlockReq\x12\x0c\n\x04\x62lob\x18\x01 \x01(\x0c\"%\n\x14SubmitMinedBlockResp\x12\r\n\x05\x65rror\x18\x01 \x01(\x08\x32\xc7\x02\n\tMiningAPI\x12_\n\x18GetBlockMiningCompatible\x12 .qrl.GetBlockMiningCompatibleReq\x1a!.qrl.GetBlockMiningCompatibleResp\x12M\n\x12GetLastBlockHeader\x12\x1a.qrl.GetLastBlockHeaderReq\x1a\x1b.qrl.GetLastBlockHeaderResp\x12\x41\n\x0eGetBlockToMine\x12\x16.qrl.GetBlockToMineReq\x1a\x17.qrl.GetBlockToMineResp\x12G\n\x10SubmitMinedBlock\x12\x18.qrl.SubmitMinedBlockReq\x1a\x19.qrl.SubmitMinedBlockRespb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'qbitmining_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_GETBLOCKMININGCOMPATIBLEREQ']._serialized_start=37
  _globals['_GETBLOCKMININGCOMPATIBLEREQ']._serialized_end=82
  _globals['_GETLASTBLOCKHEADERREQ']._serialized_start=84
  _globals['_GETLASTBLOCKHEADERREQ']._serialized_end=123
  _globals['_GETBLOCKMININGCOMPATIBLERESP']._serialized_start=125
  _globals['_GETBLOCKMININGCOMPATIBLERESP']._serialized_end=237
  _globals['_GETLASTBLOCKHEADERRESP']._serialized_start=239
  _globals['_GETLASTBLOCKHEADERRESP']._serialized_end=363
  _globals['_GETBLOCKTOMINEREQ']._serialized_start=365
  _globals['_GETBLOCKTOMINEREQ']._serialized_end=408
  _globals['_GETBLOCKTOMINERESP']._serialized_start=411
  _globals['_GETBLOCKTOMINERESP']._serialized_end=539
  _globals['_SUBMITMINEDBLOCKREQ']._serialized_start=541
  _globals['_SUBMITMINEDBLOCKREQ']._serialized_end=576
  _globals['_SUBMITMINEDBLOCKRESP']._serialized_start=578
  _globals['_SUBMITMINEDBLOCKRESP']._serialized_end=615
  _globals['_MININGAPI']._serialized_start=618
  _globals['_MININGAPI']._serialized_end=945
# @@protoc_insertion_point(module_scope)



from pyqrllib.pyqrllib import bin2hstr

from qbitcoin.core.State import State
from qbitcoin.core.StateContainer import StateContainer
from qbitcoin.core.OptimizedAddressState import OptimizedAddressState
from qbitcoin.core.misc import logger
from qbitcoin.core.txs.Transaction import Transaction
from qbitcoin.generated.qbit_pb2 import TokenBalance


class TransferTokenTransaction(Transaction):
    """
    TransferTokenTransaction for the transaction of pre-existing Token from one wallet to another.
    """

    def __init__(self, protobuf_transaction=None):
        super(TransferTokenTransaction, self).__init__(protobuf_transaction)

    @property
    def token_txhash(self):
        return self._data.transfer_token.token_txhash

    @property
    def addrs_to(self):
        return self._data.transfer_token.addrs_to

    @property
    def total_amount(self):
        total_amount = 0
        for amount in self.amounts:
            total_amount += amount

        return total_amount

    @property
    def amounts(self):
        return self._data.transfer_token.amounts

    def get_data_bytes(self):
        data_bytes = (self.master_addr +
                      self.fee.to_bytes(8, byteorder='big', signed=False) +
                      self.token_txhash)

        for index in range(0, len(self.addrs_to)):
            data_bytes = (data_bytes +
                          self.addrs_to[index] +
                          self.amounts[index].to_bytes(8, byteorder='big', signed=False))

        return data_bytes

    @staticmethod
    def create(token_txhash: bytes,
               addrs_to: list,
               amounts: list,
               fee: int,
               xmss_pk: bytes,
               master_addr: bytes = None):
        transaction = TransferTokenTransaction()

        if master_addr:
            transaction._data.master_addr = master_addr

        transaction._data.public_key = bytes(xmss_pk)

        transaction._data.transfer_token.token_txhash = token_txhash

        for addr_to in addrs_to:
            transaction._data.transfer_token.addrs_to.append(addr_to)

        for amount in amounts:
            transaction._data.transfer_token.amounts.append(amount)

        transaction._data.fee = int(fee)

        transaction.validate_or_raise(verify_signature=False)

        return transaction

    def _validate_custom(self):
        for amount in self.amounts:
            if amount == 0:
                logger.warning('Amount cannot be 0 - %s', self.amounts)
                logger.warning('TransferTokenTransaction')
                return False

        if self.fee < 0:
            logger.info('TransferTokenTransaction [%s] Invalid Fee = %d', bin2hstr(self.txhash), self.fee)
            return False

        if len(self.addrs_to) != len(self.amounts):
            logger.warning('[TransferTokenTransaction] Mismatch number of addresses to & amounts')
            logger.warning('>> Length of addresses_to %s', len(self.addrs_to))
            logger.warning('>> Length of amounts %s', len(self.amounts))
            return False

        if not OptimizedAddressState.address_is_valid(self.addr_from):
            logger.warning('[TransferTokenTransaction] Invalid address addr_from: %s', bin2hstr(self.addr_from))
            return False

        for addr_to in self.addrs_to:
            if not OptimizedAddressState.address_is_valid(addr_to):
                logger.warning('[TransferTokenTransaction] Invalid address addr_to: %s', bin2hstr(addr_to))
                return False

        return True

    # checks new tx validity based upon node statedb and node mempool.
    def _validate_extended(self, state_container: StateContainer):
        if (len(self.addrs_to) > state_container.current_dev_config.transaction_multi_output_limit or
                len(self.amounts) > state_container.current_dev_config.transaction_multi_output_limit):
            logger.warning('[TransferTokenTransaction] Number of addresses or amounts exceeds max limit')
            logger.warning('Number of addresses %s', len(self.addrs_to))
            logger.warning('Number of amounts %s', len(self.amounts))
            return False

        if len(self.addrs_to) == 0:
            logger.warning("[TransferTokenTransaction] No Addrs To found")
            return False

        tx_balance = state_container.addresses_state[self.addr_from].balance
        total_amount = self.total_amount
        if self.fee < 0 or total_amount < 0:
            logger.info('[TransferTokenTransaction] State validation failed for %s because: ', bin2hstr(self.txhash))
            logger.info('Txn amount: %s, Fee: %s', total_amount, self.fee)
            return False

        if tx_balance < self.fee:
            logger.info('[TransferTokenTransaction] State validation failed for %s because: Insufficient funds',
                        bin2hstr(self.txhash))
            logger.info('balance: %s, Fee: %s', tx_balance, self.fee)
            return False

        if (self.addr_from, self.token_txhash) not in state_container.tokens.data:
            logger.info('%s doesnt own any such token %s ', bin2hstr(self.addr_from), bin2hstr(self.token_txhash))
            return False

        token_balance = state_container.tokens.data[(self.addr_from, self.token_txhash)]
        if token_balance.balance < total_amount:
            logger.info('Insufficient amount of token')
            logger.info('Token Balance: %s, Sent Token Amount: %s',
                        token_balance.balance,
                        total_amount)
            return False

        return True

    def set_affected_address(self, addresses_set: set):
        super().set_affected_address(addresses_set)
        for addr_to in self.addrs_to:
            addresses_set.add(addr_to)

    def apply(self,
              state: State,
              state_container: StateContainer) -> bool:
        state_container.tokens.data[(self.addr_from, self.token_txhash)].balance -= self.total_amount
        decimals = state_container.tokens.data[(self.addr_from, self.token_txhash)].decimals
        address_state = state_container.addresses_state[self.addr_from]
        address_state.update_balance(state_container, self.fee, subtract=True)
        state_container.paginated_tx_hash.insert(address_state, self.txhash)

        for index in range(0, len(self.addrs_to)):
            addr_to = self.addrs_to[index]
            amount = self.amounts[index]
            address_state = state_container.addresses_state[addr_to]

            # If receiver doesn't have this token before, then initialize token balance data into state
            # before adding the new balance.
            if (addr_to, self.token_txhash) not in state_container.tokens.data:
                state_container.tokens.data[(addr_to,
                                             self.token_txhash)] = TokenBalance(balance=0,
                                                                                decimals=decimals,
                                                                                tx_hash=self.txhash,
                                                                                delete=False)
                state_container.paginated_tokens_hash.insert(address_state, self.token_txhash)

            state_container.tokens.data[(addr_to, self.token_txhash)].balance += amount

            if self.addr_from != addr_to:
                state_container.paginated_tx_hash.insert(address_state, self.txhash)

        return self._apply_state_changes_for_PK(state_container)

    def revert(self,
               state: State,
               state_container: StateContainer) -> bool:
        for index in range(0, len(self.addrs_to)):
            addr_to = self.addrs_to[index]
            amount = self.amounts[index]
            address_state = state_container.addresses_state[addr_to]
            key = (addr_to, self.token_txhash)

            state_container.tokens.data[key].balance -= amount
            # There is a chance that same address is transmitted with token multiple times,
            # in such a case, to avoid removal of token_txhash from paginated_tokens_hash
            # delete must be checked for false
            if state_container.tokens.data[key].tx_hash == self.txhash and \
                    state_container.tokens.data[key].delete is False:
                state_container.tokens.data[key].delete = True
                state_container.paginated_tokens_hash.remove(address_state, self.token_txhash)

            if self.addr_from != addr_to:
                state_container.paginated_tx_hash.remove(address_state, self.txhash)

        state_container.tokens.data[(self.addr_from, self.token_txhash)].balance += self.total_amount
        state_container.tokens.data[(self.addr_from, self.token_txhash)].delete = False
        address_state = state_container.addresses_state[self.addr_from]
        address_state.update_balance(state_container, self.fee)
        state_container.paginated_tx_hash.remove(address_state, self.txhash)

        return self._revert_state_changes_for_PK(state_container)
